"""Zeta Attention init file"""

from zeta.nn.attention.agent_attn import AgentSelfAttention
from zeta.nn.attention.attend import Attend, Intermediates
from zeta.nn.attention.cross_attn_images import MultiModalCrossAttention
from zeta.nn.attention.flash_attention import FlashAttention
from zeta.nn.attention.linear_attention import LinearAttentionVision
from zeta.nn.attention.linear_attn_l import LinearAttention
from zeta.nn.attention.local_attention import LocalAttention
from zeta.nn.attention.local_attention_mha import LocalMHA
from zeta.nn.attention.mixture_attention import (
    MixtureOfAttention,
    MixtureOfAutoregressiveAttention,
)
from zeta.nn.attention.multi_modal_causal_attention import (
    MultiModalCausalAttention,
    SimpleMMCA,
)
from zeta.nn.attention.multihead_attention import MultiheadAttention
from zeta.nn.attention.multiquery_attention import MultiQueryAttention
from zeta.nn.attention.sparse_attention import SparseAttention
from zeta.nn.attention.spatial_linear_attention import SpatialLinearAttention
from zeta.structs.transformer import Attention, AttentionLayers
from zeta.nn.attention.multi_grouped_attn import MultiGroupedQueryAttn
from zeta.nn.attention.scalable_img_self_attn import ScalableImgSelfAttention
from zeta.nn.attention.linearized_attention import LinearizedAttention


__all__ = [
    "Attend",
    "FlashAttention",
    "LocalAttention",
    "LocalMHA",
    "Intermediates",
    "MixtureOfAttention",
    "MixtureOfAutoregressiveAttention",
    "MultiModalCausalAttention",
    "SimpleMMCA",
    "MultiheadAttention",
    "MultiQueryAttention",
    "MultiModalCrossAttention",
    "SparseAttention",
    "SpatialLinearAttention",
    "LinearAttentionVision",
    "AgentSelfAttention",
    "LinearAttention",
    "Attention",
    "AttentionLayers",
    "MultiGroupedQueryAttn",
    "ScalableImgSelfAttention",
    "LinearizedAttention",
]



import torch
from einops import rearrange
from torch import nn


class DynamicPositionBias(nn.Module):
    """
    Dynamic Position Bias

    Args:
        dim (int): dimension of the embedding
        heads (int): number of heads

    Attributes:
        mlp (nn.Sequential): multi-layer perceptron

    Usage:
    >>> X = torch.randn(1, 10, 512)
    >>> bias = DynamicPositionBias(dim=512, heads=8)

    """

    def __init__(self, dim, heads):
        super().__init__()
        self.mlp = nn.Sequential(
            nn.Linear(1, dim),
            nn.SiLU(),
            nn.Linear(dim, dim),
            nn.SiLU(),
            nn.Linear(dim, heads),
        )

    @property
    def device(self):
        """device property"""
        return next(self.parameters()).device

    def forward(self, i, j):
        """Forward method of DynamicPositionBias"""
        device = self.device
        assert j >= i

        rel_dist = torch.arange(j, dtype=torch.float, device=device)
        bias = self.mlp(rearrange(rel_dist, "... -> ... 1"))

        i_seq = torch.arange(j - i, j, device=device)
        j_seq = torch.arange(j, device=device)

        rel_dist_indices = (
            rearrange(i_seq, "i -> i 1") - rearrange(j_seq, "j -> 1 j")
        ).abs()

        bias = rearrange(bias[rel_dist_indices], "i j h -> h i j")
        return bias



import torch
import torch.nn.functional as F
import math
from typing import List


class KANLinear(torch.nn.Module):
    def __init__(
        self,
        in_features: int,
        out_features: int,
        grid_size: int = 5,
        spline_order: int = 3,
        scale_noise: float = 0.1,
        scale_base: float = 1.0,
        scale_spline: float = 1.0,
        enable_standalone_scale_spline: bool = True,
        base_activation: torch.nn.Module = torch.nn.SiLU,
        grid_eps: float = 0.02,
        grid_range: List[float] = [-1, 1],
    ):
        super(KANLinear, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.grid_size = grid_size
        self.spline_order = spline_order

        h = (grid_range[1] - grid_range[0]) / grid_size
        grid = (
            (
                torch.arange(-spline_order, grid_size + spline_order + 1) * h
                + grid_range[0]
            )
            .expand(in_features, -1)
            .contiguous()
        )
        self.register_buffer("grid", grid)

        self.base_weight = torch.nn.Parameter(
            torch.Tensor(out_features, in_features)
        )
        self.spline_weight = torch.nn.Parameter(
            torch.Tensor(out_features, in_features, grid_size + spline_order)
        )
        if enable_standalone_scale_spline:
            self.spline_scaler = torch.nn.Parameter(
                torch.Tensor(out_features, in_features)
            )

        self.scale_noise = scale_noise
        self.scale_base = scale_base
        self.scale_spline = scale_spline
        self.enable_standalone_scale_spline = enable_standalone_scale_spline
        self.base_activation = base_activation()
        self.grid_eps = grid_eps

        self.reset_parameters()

    def reset_parameters(self):
        torch.nn.init.kaiming_uniform_(
            self.base_weight, a=math.sqrt(5) * self.scale_base
        )
        with torch.no_grad():
            noise = (
                (
                    torch.rand(
                        self.grid_size + 1, self.in_features, self.out_features
                    )
                    - 1 / 2
                )
                * self.scale_noise
                / self.grid_size
            )
            self.spline_weight.data.copy_(
                (
                    self.scale_spline
                    if not self.enable_standalone_scale_spline
                    else 1.0
                )
                * self.curve2coeff(
                    self.grid.T[self.spline_order : -self.spline_order],
                    noise,
                )
            )
            if self.enable_standalone_scale_spline:
                # torch.nn.init.constant_(self.spline_scaler, self.scale_spline)
                torch.nn.init.kaiming_uniform_(
                    self.spline_scaler, a=math.sqrt(5) * self.scale_spline
                )

    def b_splines(self, x: torch.Tensor):
        """
        Compute the B-spline bases for the given input tensor.

        Args:
            x (torch.Tensor): Input tensor of shape (batch_size, in_features).

        Returns:
            torch.Tensor: B-spline bases tensor of shape (batch_size, in_features, grid_size + spline_order).
        """
        assert x.dim() == 2 and x.size(1) == self.in_features

        grid: torch.Tensor = (
            self.grid
        )  # (in_features, grid_size + 2 * spline_order + 1)
        x = x.unsqueeze(-1)
        bases = ((x >= grid[:, :-1]) & (x < grid[:, 1:])).to(x.dtype)
        for k in range(1, self.spline_order + 1):
            bases = (
                (x - grid[:, : -(k + 1)])
                / (grid[:, k:-1] - grid[:, : -(k + 1)])
                * bases[:, :, :-1]
            ) + (
                (grid[:, k + 1 :] - x)
                / (grid[:, k + 1 :] - grid[:, 1:(-k)])
                * bases[:, :, 1:]
            )

        assert bases.size() == (
            x.size(0),
            self.in_features,
            self.grid_size + self.spline_order,
        )
        return bases.contiguous()

    def curve2coeff(self, x: torch.Tensor, y: torch.Tensor):
        """
        Compute the coefficients of the curve that interpolates the given points.

        Args:
            x (torch.Tensor): Input tensor of shape (batch_size, in_features).
            y (torch.Tensor): Output tensor of shape (batch_size, in_features, out_features).

        Returns:
            torch.Tensor: Coefficients tensor of shape (out_features, in_features, grid_size + spline_order).
        """
        assert x.dim() == 2 and x.size(1) == self.in_features
        assert y.size() == (x.size(0), self.in_features, self.out_features)

        A = self.b_splines(x).transpose(
            0, 1
        )  # (in_features, batch_size, grid_size + spline_order)
        B = y.transpose(0, 1)  # (in_features, batch_size, out_features)
        solution = torch.linalg.lstsq(
            A, B
        ).solution  # (in_features, grid_size + spline_order, out_features)
        result = solution.permute(
            2, 0, 1
        )  # (out_features, in_features, grid_size + spline_order)

        assert result.size() == (
            self.out_features,
            self.in_features,
            self.grid_size + self.spline_order,
        )
        return result.contiguous()

    @property
    def scaled_spline_weight(self):
        return self.spline_weight * (
            self.spline_scaler.unsqueeze(-1)
            if self.enable_standalone_scale_spline
            else 1.0
        )

    def forward(self, x: torch.Tensor):
        assert x.dim() == 2 and x.size(1) == self.in_features

        base_output = F.linear(self.base_activation(x), self.base_weight)
        spline_output = F.linear(
            self.b_splines(x).view(x.size(0), -1),
            self.scaled_spline_weight.view(self.out_features, -1),
        )
        return base_output + spline_output

    @torch.no_grad()
    def update_grid(self, x: torch.Tensor, margin=0.01):
        assert x.dim() == 2 and x.size(1) == self.in_features
        batch = x.size(0)

        splines = self.b_splines(x)  # (batch, in, coeff)
        splines = splines.permute(1, 0, 2)  # (in, batch, coeff)
        orig_coeff = self.scaled_spline_weight  # (out, in, coeff)
        orig_coeff = orig_coeff.permute(1, 2, 0)  # (in, coeff, out)
        unreduced_spline_output = torch.bmm(
            splines, orig_coeff
        )  # (in, batch, out)
        unreduced_spline_output = unreduced_spline_output.permute(
            1, 0, 2
        )  # (batch, in, out)

        # sort each channel individually to collect data distribution
        x_sorted = torch.sort(x, dim=0)[0]
        grid_adaptive = x_sorted[
            torch.linspace(
                0,
                batch - 1,
                self.grid_size + 1,
                dtype=torch.int64,
                device=x.device,
            )
        ]

        uniform_step = (
            x_sorted[-1] - x_sorted[0] + 2 * margin
        ) / self.grid_size
        grid_uniform = (
            torch.arange(
                self.grid_size + 1, dtype=torch.float32, device=x.device
            ).unsqueeze(1)
            * uniform_step
            + x_sorted[0]
            - margin
        )

        grid = (
            self.grid_eps * grid_uniform + (1 - self.grid_eps) * grid_adaptive
        )
        grid = torch.concatenate(
            [
                grid[:1]
                - uniform_step
                * torch.arange(
                    self.spline_order, 0, -1, device=x.device
                ).unsqueeze(1),
                grid,
                grid[-1:]
                + uniform_step
                * torch.arange(
                    1, self.spline_order + 1, device=x.device
                ).unsqueeze(1),
            ],
            dim=0,
        )

        self.grid.copy_(grid.T)
        self.spline_weight.data.copy_(
            self.curve2coeff(x, unreduced_spline_output)
        )

    def regularization_loss(
        self, regularize_activation=1.0, regularize_entropy=1.0
    ):
        """
        Compute the regularization loss.

        This is a dumb simulation of the original L1 regularization as stated in the
        paper, since the original one requires computing absolutes and entropy from the
        expanded (batch, in_features, out_features) intermediate tensor, which is hidden
        behind the F.linear function if we want an memory efficient implementation.

        The L1 regularization is now computed as mean absolute value of the spline
        weights. The authors implementation also includes this term in addition to the
        sample-based regularization.
        """
        l1_fake = self.spline_weight.abs().mean(-1)
        regularization_loss_activation = l1_fake.sum()
        p = l1_fake / regularization_loss_activation
        regularization_loss_entropy = -torch.sum(p * p.log())
        return (
            regularize_activation * regularization_loss_activation
            + regularize_entropy * regularization_loss_entropy
        )


class KAN(torch.nn.Module):
    """
    KAN (Kernel Activation Network) module.

    Args:
        layers_hidden (list): List of integers representing the number of hidden units in each layer.
        grid_size (int, optional): Size of the grid. Defaults to 5.
        spline_order (int, optional): Order of the spline. Defaults to 3.
        scale_noise (float, optional): Scale factor for the noise. Defaults to 0.1.
        scale_base (float, optional): Scale factor for the base. Defaults to 1.0.
        scale_spline (float, optional): Scale factor for the spline. Defaults to 1.0.
        base_activation (torch.nn.Module, optional): Activation function for the base. Defaults to torch.nn.SiLU.
        grid_eps (float, optional): Epsilon value for the grid. Defaults to 0.02.
        grid_range (list, optional): Range of the grid. Defaults to [-1, 1].

    Example:
    >>> kan = KAN([2, 3, 1])
    >>> x = torch.randn(10, 2)
    >>> y = kan(x)

    """

    def __init__(
        self,
        layers_hidden: List[int],
        grid_size: int = 5,
        spline_order: int = 3,
        scale_noise: float = 0.1,
        scale_base: float = 1.0,
        scale_spline: float = 1.0,
        base_activation: torch.nn.Module = torch.nn.SiLU,
        grid_eps: float = 0.02,
        grid_range: List[float] = [-1, 1],
    ) -> None:
        super(KAN, self).__init__()
        self.grid_size = grid_size
        self.spline_order = spline_order

        self.layers = torch.nn.ModuleList()
        for in_features, out_features in zip(layers_hidden, layers_hidden[1:]):
            self.layers.append(
                KANLinear(
                    in_features,
                    out_features,
                    grid_size=grid_size,
                    spline_order=spline_order,
                    scale_noise=scale_noise,
                    scale_base=scale_base,
                    scale_spline=scale_spline,
                    base_activation=base_activation,
                    grid_eps=grid_eps,
                    grid_range=grid_range,
                )
            )

    def forward(self, x: torch.Tensor, update_grid=False):
        """
        Forward pass of the KAN module.

        Args:
            x (torch.Tensor): Input tensor.
            update_grid (bool, optional): Whether to update the grid. Defaults to False.

        Returns:
            torch.Tensor: Output tensor.
        """
        for layer in self.layers:
            if update_grid:
                layer.update_grid(x)
            x = layer(x)
        return x

    def regularization_loss(
        self, regularize_activation=1.0, regularize_entropy=1.0
    ):
        """
        Compute the regularization loss of the KAN module.

        Args:
            regularize_activation (float, optional): Regularization factor for activation. Defaults to 1.0.
            regularize_entropy (float, optional): Regularization factor for entropy. Defaults to 1.0.

        Returns:
            torch.Tensor: Regularization loss.
        """
        return sum(
            layer.regularization_loss(regularize_activation, regularize_entropy)
            for layer in self.layers
        )


# x = torch.randn(2, 3, 1)
# kan = KAN(
#     layers_hidden=[2, 3, 1],
# )
# y = kan(x)
# print(y)



import torch
from einops import rearrange, repeat
from einops.layers.torch import Rearrange
from torch import einsum, nn
from torch.nn import Module

# functions


def exists(v):
    return v is not None


# main class


class AgentSelfAttention(Module):
    """
    Self-attention module for agent tokens in a neural network.

    Args:
        dim (int): The input dimension.
        num_agent_tokens (int): The number of agent tokens.
        dim_head (int, optional): The dimension of each attention head. Defaults to 64.
        heads (int, optional): The number of attention heads. Defaults to 8.
        dropout (float, optional): The dropout rate. Defaults to 0.0.
        talking_heads (bool, optional): Whether to use talking heads mechanism. Defaults to True.
        gate (bool, optional): Whether to apply gating mechanism. Defaults to True.
        combine_agent_tokens (bool, optional): Whether to combine agent tokens. Defaults to False.

    Examples::
        >>> import torch
        >>> from zeta.nn.attention import AgentSelfAttention
        >>> agent_self_attn = AgentSelfAttention(dim=64, num_agent_tokens=16)
        >>> x = torch.randn(2, 64)
        >>> output = agent_self_attn(x)
        >>> output.shape
        torch.Size([2, 64])
    """

    def __init__(
        self,
        dim,
        *,
        num_agent_tokens,
        dim_head=64,
        heads=8,
        dropout=0.0,
        talking_heads=True,
        gate=True,
        combine_agent_tokens=False,
    ):
        super().__init__()
        self.scale = dim_head**-0.5
        dim_inner = dim_head * heads

        self.to_qkv = nn.Sequential(
            nn.Linear(dim, dim_inner * 3, bias=False),
            Rearrange("b n (qkv h d) -> qkv b h n d", h=heads, qkv=3),
        )

        self.to_gates = (
            nn.Sequential(
                nn.Linear(dim, heads),
                Rearrange("b n h -> b h n 1"),
                nn.Sigmoid(),
            )
            if gate
            else None
        )

        self.agent_tokens = nn.Parameter(
            torch.zeros(heads, num_agent_tokens, dim_head)
        )
        nn.init.normal_(self.agent_tokens, std=0.02)

        self.qa_talking_heads = (
            nn.Conv2d(heads, heads, 1, bias=False)
            if talking_heads
            else nn.Identity()
        )
        self.ak_talking_heads = (
            nn.Conv2d(heads, heads, 1, bias=False)
            if talking_heads
            else nn.Identity()
        )

        self.qa_dropout = nn.Dropout(dropout)
        self.ak_dropout = nn.Dropout(dropout)

        self.to_out = nn.Sequential(
            Rearrange("b h n d -> b n (h d)"),
            nn.Linear(dim_inner, dim, bias=False),
        )

    def forward(
        self, x, mask=None, agent_tokens=None, return_agent_tokens=False
    ):
        batch = x.shape[0]

        q, k, v = self.to_qkv(x)

        if exists(agent_tokens):
            a = agent_tokens
        else:
            a = repeat(self.agent_tokens, "h m d -> b h m d", b=batch)

        a = a * self.scale

        qa_sim = einsum("b h i d, b h j d -> b h i j", q, a)
        ak_sim = einsum("b h i d, b h j d -> b h i j", a, k)

        if exists(mask):
            max_neg_value = -torch.finfo(qa_sim.dtype).max
            ak_sim = ak_sim.masked_fill(
                ~rearrange(mask, "b j -> b 1 1 j"), max_neg_value
            )

        qa_attn = qa_sim.softmax(dim=-1)
        ak_attn = ak_sim.softmax(dim=-1)

        qa_attn = self.qa_dropout(qa_attn)
        ak_attn = self.ak_dropout(ak_attn)

        qa_attn = self.qa_talking_heads(qa_attn)
        ak_attn = self.ak_talking_heads(ak_attn)

        agent_gathered_tokens = einsum(
            "b h i j, b h j d -> b h i d", ak_attn, v
        )

        out = einsum(
            "b h i j, b h j d -> b h i d", qa_attn, agent_gathered_tokens
        )

        if exists(mask):
            out = out.masked_fill(~rearrange(mask, "b n -> b 1 n 1"), 0.0)

        if exists(self.to_gates):
            out = out * self.to_gates(x)

        out = self.to_out(out)

        if not return_agent_tokens:
            return out

        return out, agent_gathered_tokens



import torch
from einops import rearrange, reduce
from torch import nn


class DropSample(nn.Module):
    def __init__(self, prob=0):
        super().__init__()
        self.prob = prob

    def forward(self, x):
        device = x.device

        if self.prob == 0.0 or (not self.training):
            return x

        keep_mask = (
            torch.FloatTensor((x.shape[0], 1, 1, 1), device=device).uniform_()
            > self.prob
        )
        return x + keep_mask / (1 - self.prob)


class SqueezeExcitation(nn.Module):
    """
    Squeeze-and-Excitation module for channel-wise feature recalibration.

    Args:
        dim (int): Number of input channels.
        shrinkage_rate (float, optional): Shrinkage rate for the hidden dimension. Defaults to 0.25.
    """

    def __init__(self, dim, shrinkage_rate=0.25):
        super().__init__()
        hidden_dim = int(dim * shrinkage_rate)

        self.gate = nn.Sequential(
            nn.Linear(dim, hidden_dim, bias=False),
            nn.SiLU(),
            nn.Linear(hidden_dim, dim, bias=False),
            nn.Sigmoid(),
        )

    def forward(self, x):
        b, c, h, w = x.shape
        y = reduce(x, "b c h w -> b c", "mean")
        y = self.gate(y)
        y = rearrange(y, "b c -> b c () ()")
        return x * y.expand_as(x)


class MBConvResidual(nn.Module):
    def __init__(self, fn, dropout=0.0):
        super().__init__()
        self.fn = fn
        self.downsample = DropSample(dropout)

    def forward(self, x):
        out = self.fn(x)
        out = self.dropsample(out)
        return out + x


def MBConv(
    dim_in,
    dim_out,
    *,
    downsample,
    expansion_rate=4,
    shrinkage_rate=0.25,
    dropout=0.0,
):
    """
    MobileNetV3 Bottleneck Convolution (MBConv) block.

    Args:
        dim_in (int): Number of input channels.
        dim_out (int): Number of output channels.
        downsample (bool): Whether to downsample the spatial dimensions.
        expansion_rate (float, optional): Expansion rate for the hidden dimension. Defaults to 4.
        shrinkage_rate (float, optional): Shrinkage rate for the squeeze excitation. Defaults to 0.25.
        dropout (float, optional): Dropout rate. Defaults to 0.0.

    Returns:
        nn.Sequential: MBConv block.
    """
    hidden_dim = int(expansion_rate * dim_out)
    stride = 2 if downsample else 1

    net = nn.Sequential(
        nn.Conv2d(dim_in, hidden_dim, 1),
        nn.BatchNorm2d(hidden_dim),
        nn.GELU(),
        nn.Conv2d(
            hidden_dim,
            hidden_dim,
            3,
            stride=stride,
            padding=1,
            groups=hidden_dim,
        ),
        nn.BatchNorm2d(hidden_dim),
        nn.GELU(),
        SqueezeExcitation(hidden_dim, shrinkage_rate=shrinkage_rate),
        nn.Conv2d(hidden_dim, dim_out, 1),
        nn.BatchNorm2d(dim_out),
    )

    if dim_in == dim_out and not downsample:
        net = MBConvResidual(net, dropout=dropout)

    return net



"""
DOES NOT WORK:
    - Need to configure the input shape to match the input shape of regular text features

VisuaL Expert module from: https://arxiv.org/pdf/2311.03079.pdf

Visual expert module. We add a visual expert module to each layer to enable deep visual-language
feature alignment. Specifically, the visual expert module in each layer consists of a QKV matrix
and an MLP in each layer. The shapes of the QKV matrix and MLP are identical to those in the
pretrained language model and initialized from them. The motivation is that each attention head
in the language model captures a certain aspect of semantic information, while a trainable visual
expert can transform the image features to align with the different heads, therefore enabling deep
fusion.

Formally, suppose that the input hidden states of an attention layer are X ∈ R
B×H×(LI+LT )×D,
where B is the batch size, LI and LT are the lengths of image and text sequences, H is the number
of attention heads, and D is the hidden size. In the attention with visual expert, X is first split as
4

Shape = B, SEQ_LEN, DIM or regular text shape
"""

import torch
from torch import nn

from zeta.nn.attention.multihead_attention import MultiheadAttention
from zeta.nn.modules.simple_feedforward import SimpleFeedForward


class VisualExpert:
    """
    Visual Expert from https://arxiv.org/pdf/2311.03079.pdf

    Visual expert module. We add a visual expert module to each layer to enable deep visual-language
    feature alignment. Specifically, the visual expert module in each layer consists of a QKV matrix
    and an MLP in each layer. The shapes of the QKV matrix and MLP are identical to those in the
    pretrained language model and initialized from them. The motivation is that each attention head
    in the language model captures a certain aspect of semantic information, while a trainable visual
    expert can transform the image features to align with the different heads, therefore enabling deep
    fusion.

    Args:
        dim (int): The dimension of the input features.
        hidden_dim (int): The dimension of the hidden layer in the feedforward.
        dropout (float): The dropout rate.
        heads (int): The number of heads in the multihead attention.

    Attributes:
        dim (int): The dimension of the input features.
        hidden_dim (int): The dimension of the hidden layer in the feedforward.
        dropout (float): The dropout rate.
        heads (int): The number of heads in the multihead attention.
        norm (nn.LayerNorm): The layer norm.
        q_proj (nn.Linear): The projection of the query.
        k_proj (nn.Linear): The projection of the key.
        v_proj (nn.Linear): The projection of the value.
        attention (MultiheadAttention): The multihead attention.
        feedforward (SimpleFeedForward): The feedforward.

    Input shape: (B, SEQ_LEN, DIM) or regular text shape

    Output shape: (B, SEQ_LEN, DIM) or regular text shape

    Example:
        >>> visual_expert = VisualExpert(1024, 2048, 0.1, 16)
        >>> x = torch.randn(1, 10, 1024)
        >>> out = visual_expert(x)
        >>> out.shape
        torch.Size([1, 10, 1024])

    """

    def __init__(
        self,
        dim: int,
        hidden_dim: int,
        dropout: float,
        heads: int,
    ):
        self.dim = dim
        self.hidden_dim = hidden_dim
        self.dropout = dropout
        self.heads = heads

        # Normalization
        self.norm = nn.LayerNorm(dim)

        # Projections
        self.q_proj = nn.Linear(dim, dim)
        self.k_proj = nn.Linear(dim, dim)
        self.v_proj = nn.Linear(dim, dim)

        # Attention
        self.attention = MultiheadAttention(dim, heads, dropout)

        # Feedforward
        self.feedforward = SimpleFeedForward(dim, hidden_dim, dropout)

    def __call__(self, x: torch.Tensor):
        """Forward pass as shown in the diagram"""

        # Apply Layernorm first
        normalized = self.norm(x)

        # Split into text and image features
        x_text = normalized
        x_image = normalized

        # Apply QKV projections for text
        q_text, k_text, v_text = (
            self.q_proj(x_text),
            self.k_proj(x_text),
            self.v_proj(x_text),
        )

        # Apply QKV projections for image
        q_img, k_img, v_img = (
            self.q_proj(x_image),
            self.k_proj(x_image),
            self.v_proj(x_image),
        )

        # Apply attention where the image features are appended infront of the text features,
        # Concat the q, k, v of text and images together
        q = torch.cat((q_text, q_img))  # , dim=-1)
        k = torch.cat((k_text, k_img))  # , dim=-1)
        v = torch.cat((v_text, v_img))  # , dim=-1)

        # Apply attention
        out = self.attention(q, k, v)

        # Add the output of the attention with the normed x
        out = out + x

        # Another Norm
        normalized = self.norm(out)

        # Seperate text and image features
        out_text = normalized
        out_image = normalized  # torch.split(normalized, self.dim)  # dim=-1)

        # Apply feedforward to both text and image features
        out_text = self.feedforward(out_text)
        out_img = self.feedforward(out_image)

        # Add the output of the feedforwards together with the output of the added attention + norm
        out = out_text + out_img + out

        return out



import torch
from einops import pack, rearrange, repeat, unpack
from torch import einsum, nn

from zeta.nn.embeddings.sinusoidal import (
    SinusoidalEmbeddings,
    apply_rotary_pos_emb,
)
from zeta.utils.main import (
    default,
    exists,
    l2norm,
    look_around,
    max_neg_values,
    pad_to_multiple,
)

# constant
TOKEN_SELF_ATTN_VALUE = -5e4


class LocalAttention(nn.Module):
    """

    The LocalAttention module provides a mechanism to perform local attention operations.
    Unlike global attention where every token can attend to every other token,
    in local attention each token can only attend to a subset of tokens within a defined window. This reduces the computational cost and captures the local structure in sequences like text or time-series data.

    Args:
        window_size: (int) The size of the attention window.
        causal: (bool, optional) If set to True, ensures causal attention. Default: False.
        look_backward: (int, optional) How many positions to look backward from the current position. Default: 1.
        look_forward: (int, optional) How many positions to look forward from the current position. Default: None which implies 0 if causal is True.
        dropout: (float, optional) Dropout rate for attention weights. Default: 0..
        shared_qk: (bool, optional) If set to True, the query and key are the same. Useful for certain types of attention mechanisms. Default: False.
        rel_pos_emb_config: (Optional) Deprecated. Configuration for the relative positional embeddings.
        dim: (int, optional) Dimension of embeddings. Only needed if rel_pos_emb_config is not provided.
        autopad: (bool, optional) If set to True, sequence will be automatically padded to be divisible by the window size. Default: False.
        exact_windowsize: (bool, optional) Ensures exact window size for non-causal attention. Default: False.
        scale: (Optional) Scaling factor for the queries.
        use_rotary_pos_emb: (bool, optional) If set to True, rotary positional embeddings will be used. Default: True.
        use_xpos: (bool, optional) If set to True, allows for extrapolation of window sizes. Requires use_rotary_pos_emb to be True. Default: False.
        xpos_scale_base: (Optional) Base scaling factor for extrapolated window sizes.

    Usage:
    >>> model = LocalAttention(64, 1, 1, 0.1)
    >>> x = torch.randn(1, 768)
    >>> model(x).shape

    """

    def __init__(
        self,
        window_size,
        causal=False,
        look_backward=1,
        look_forward=None,
        dropout=0.0,
        shared_qk=False,
        rel_pos_emb_config=None,
        dim=None,
        autopad=False,
        exact_windowsize=False,
        scale=None,
        use_rotary_pos_emb=True,
        use_xpos=False,
        xpos_scale_base=None,
    ):
        super().__init__()
        look_forward = default(look_forward, 0 if causal else 1)
        assert not (
            causal and look_forward > 0
        ), "you cannot look forward if causal"

        self.scale = scale

        self.window_size = window_size
        self.autopad = autopad
        self.exact_windowsize = exact_windowsize

        self.causal = causal

        self.look_backward = look_backward
        self.look_forward = look_forward

        self.dropout = nn.Dropout(dropout)

        self.shared_qk = shared_qk

        # relative positions

        self.rel_pos = None
        self.use_xpos = use_xpos

        if use_rotary_pos_emb and (
            exists(rel_pos_emb_config) or exists(dim)
        ):  # backwards compatible with old `rel_pos_emb_config` deprecated argument
            if exists(rel_pos_emb_config):
                dim = rel_pos_emb_config[0]

            self.rel_pos = SinusoidalEmbeddings(
                dim,
                use_xpos=use_xpos,
                scale_base=default(xpos_scale_base, window_size // 2),
            )

    """

    Forward Method
        Parameters
        q: (Tensor) The query tensor.

        k: (Tensor) The key tensor.

        v: (Tensor) The value tensor.

        mask: (Optional[Tensor]) A mask tensor for the keys. Can also be passed as input_mask.

        input_mask: (Optional[Tensor]) Another way to pass the mask tensor for keys.

        attn_bias: (Optional[Tensor]) Additional biases to add to the attention scores.

        window_size: (Optional[int]) If provided, this window size will override the default window size defined during initialization.

        Returns
        out: (Tensor) The output tensor after the attention operation.
    """

    def forward(
        self,
        q,
        k,
        v,
        mask=None,
        input_mask=None,
        attn_bias=None,
        window_size=None,
    ):
        mask = default(mask, input_mask)

        assert not (
            exists(window_size) and not self.use_xpos
        ), "cannot perform window size extrapolation if xpos is not turned on"

        (
            _shape,
            autopad,
            pad_value,
            window_size,
            causal,
            look_backward,
            look_forward,
            shared_qk,
        ) = (
            q.shape,
            self.autopad,
            -1,
            default(window_size, self.window_size),
            self.causal,
            self.look_backward,
            self.look_forward,
            self.shared_qk,
        )

        # https://github.com/arogozhnikov/einops/blob/master/docs/4-pack-and-unpack.ipynb
        (q, packed_shape), (k, _), (v, _) = map(
            lambda t: pack([t], "* n d"), (q, k, v)
        )

        # auto padding

        if autopad:
            orig_seq_len = q.shape[1]
            (needed_pad, q), (_, k), (_, v) = map(
                lambda t: pad_to_multiple(t, self.window_size, dim=-2),
                (q, k, v),
            )

        b, n, dim_head, device, _dtype = *q.shape, q.device, q.dtype

        scale = default(self.scale, dim_head**-0.5)

        assert (n % window_size) == 0, (
            f"sequence length {n} must be divisible by window size"
            f" {window_size} for local attention"
        )

        windows = n // window_size

        if shared_qk:
            k = l2norm(k)

        seq = torch.arange(n, device=device)
        b_t = rearrange(seq, "(w n) -> 1 w n", w=windows, n=window_size)

        # bucketing

        bq, bk, bv = map(
            lambda t: rearrange(t, "b (w n) d -> b w n d", w=windows), (q, k, v)
        )

        bq = bq * scale

        look_around_kwargs = dict(
            backward=look_backward, forward=look_forward, pad_value=pad_value
        )

        bk = look_around(bk, **look_around_kwargs)
        bv = look_around(bv, **look_around_kwargs)

        # rotary embeddings

        if exists(self.rel_pos):
            pos_emb, xpos_scale = self.rel_pos(bk)
            bq, bk = apply_rotary_pos_emb(bq, bk, pos_emb, scale=xpos_scale)

        # calculate positions for masking

        bq_t = b_t
        bq_k = look_around(b_t, **look_around_kwargs)

        bq_t = rearrange(bq_t, "... i -> ... i 1")
        bq_k = rearrange(bq_k, "... j -> ... 1 j")

        pad_mask = bq_k == pad_value

        sim = einsum("b h i e, b h j e -> b h i j", bq, bk)

        if exists(attn_bias):
            heads = attn_bias.shape[0]
            assert (b % heads) == 0

            attn_bias = repeat(attn_bias, "h i j -> (b h) 1 i j", b=b // heads)
            sim = sim + attn_bias

        mask_value = max_neg_values(sim)

        if shared_qk:
            self_mask = bq_t == bq_k
            sim = sim.masked_fill(self_mask, TOKEN_SELF_ATTN_VALUE)
            del self_mask

        if causal:
            causal_mask = bq_t < bq_k

            if self.exact_windowsize:
                max_causal_window_size = self.window_size * self.look_backward
                causal_mask = causal_mask | (
                    bq_t > (bq_k + max_causal_window_size)
                )

            sim = sim.masked_fill(causal_mask, mask_value)
            del causal_mask

        # masking out for exact window size for non-causal
        # as well as masking out for padding value

        if not causal and self.exact_windowsize:
            max_backward_window_size = self.window_size * self.look_backward
            max_forward_window_size = self.window_size * self.look_forward
            window_mask = (
                ((bq_k - max_forward_window_size) > bq_t)
                | (bq_t > (bq_k + max_backward_window_size))
                | pad_mask
            )
            sim = sim.masked_fill(window_mask, mask_value)
        else:
            sim = sim.masked_fill(pad_mask, mask_value)

        # take care of key padding mask passed in

        if exists(mask):
            batch = mask.shape[0]
            assert (b % batch) == 0

            h = b // mask.shape[0]

            if autopad:
                _, mask = pad_to_multiple(
                    mask, window_size, dim=-1, value=False
                )

            mask = rearrange(
                mask, "... (w n) -> (...) w n", w=windows, n=window_size
            )
            mask = look_around(
                mask, **{**look_around_kwargs, "pad_value": False}
            )
            mask = rearrange(mask, "... j -> ... 1 j")
            mask = repeat(mask, "b ... -> (b h) ...", h=h)
            sim = sim.masked_fill(~mask, mask_value)
            del mask

        # attention

        attn = sim.softmax(dim=-1)
        attn = self.dropout(attn)

        # aggregation

        out = einsum("b h i j, b h j e -> b h i e", attn, bv)
        out = rearrange(out, "b w n d -> b (w n) d")

        if autopad:
            out = out[:, :orig_seq_len, :]

        out, *_ = unpack(out, packed_shape, "* n d")
        return out



from einops.layers.torch import Rearrange
from torch import nn


class PatchEmbeddings(nn.Module):
    """
    Patch embeddings for images.

    Parameters:
    - dim_in (int): The input dimension.
    - dim_out (int): The output dimension.
    - seq_len (int): The sequence length.

    Attributes:
    - embedding (nn.Module): The embedding layer.

    Example:

        >>> module = PatchEmbeddings(3, 4, 5)
        >>> x = torch.randn(2, 3, 5, 5)
        >>> y = module(x)
        >>> y.shape
        torch.Size([2, 20, 5])

    """

    def __init__(self, dim_in, dim_out, seq_len):
        super().__init__()
        self.embedding = nn.Sequential(
            Rearrange("... rd -> ... (r d)"),
            nn.LayerNorm(seq_len * dim_in),
            nn.Linear(seq_len * dim_in, dim_out),
            nn.LayerNorm(dim_out),
        )

    def forward(self, x):
        return self.embedding(x)



from torch import nn


def text_to_tensor(
    text: str,
    tokenizer: callable,
    process_func: callable,
    dim: int,
    num_tokens: int,
):
    """
    Converts a given text into a tensor representation.

    Args:
        text (str): The input text to be converted.
        tokenizer (callable): A callable object that tokenizes the text.
        process_func (callable): A callable object that processes the tokens.
        dim (int): The dimension of the embedding.
        num_tokens (int): The number of tokens in the vocabulary.

    Returns:
        out: The tensor representation of the input text.
    """
    tokens = tokenizer(text)

    # Truncate or pad the tokens to the specified length
    tokens = process_func(tokens)

    # Convert the tokens to a tensor
    out = nn.Embedding(num_tokens, dim)(tokens)
    return out



from zeta.nn.embeddings.abc_pos_emb import AbsolutePositionalEmbedding
from zeta.nn.embeddings.embedding import BaseEmbedding, Embedding, TextEmbedding
from zeta.nn.embeddings.multiway_network import (
    MultiwayEmbedding,
    MultiwayNetwork,
    MultiwayWrapper,
    set_split_position,
)
from zeta.nn.embeddings.nominal_embeddings import NominalEmbedding
from zeta.nn.embeddings.positional import PositionalEmbedding
from zeta.nn.embeddings.positional_interpolation import (
    PositionInterpolationEmbeddings,
)
from zeta.nn.embeddings.qfsp_embeddings import QFTSPEmbedding
from zeta.nn.embeddings.qft_embeddings import QFTSPEmbeddings
from zeta.nn.embeddings.rope import RotaryEmbedding
from zeta.nn.embeddings.sine_positional import SinePositionalEmbedding
from zeta.nn.embeddings.sinusoidal import SinusoidalEmbeddings
from zeta.nn.embeddings.truncated_rope import TruncatedRotaryEmbedding
from zeta.nn.embeddings.vis_lang_emb import VisionLanguageEmbedding
from zeta.nn.embeddings.vision_emb import VisionEmbedding
from zeta.nn.embeddings.xpos_relative_position import (
    XPOS,
    apply_rotary_pos_emb,
    duplicate_interleave,
    fixed_pos_embedding,
    rotate_every_two,
)
from zeta.nn.embeddings.yarn import YarnEmbedding
from zeta.nn.embeddings.scaled_sinusoidal_embeddings import (
    ScaledSinusoidalEmbedding,
)
from zeta.nn.embeddings.mi_rope import MIRoPE


__all__ = [
    "AbsolutePositionalEmbedding",
    "BaseEmbedding",
    "Embedding",
    "TextEmbedding",
    "MultiwayEmbedding",
    "MultiwayNetwork",
    "NominalEmbedding",
    "PositionalEmbedding",
    "PositionInterpolationEmbeddings",
    "RotaryEmbedding",
    "SinusoidalEmbeddings",
    "TruncatedRotaryEmbedding",
    "VisionLanguageEmbedding",
    "XPOS",
    "apply_rotary_pos_emb",
    "rotate_every_two",
    "YarnEmbedding",
    "SinePositionalEmbedding",
    "QFTSPEmbeddings",
    "QFTSPEmbedding",
    "set_split_position",
    "MultiwayWrapper",
    "MultiwayNetwork",
    "MultiwayEmbedding",
    "fixed_pos_embedding",
    "duplicate_interleave",
    "VisionEmbedding",
    "ScaledSinusoidalEmbedding",
    "MIRoPE",
]



import torch.nn.functional as F
from torch import nn


class ResidualBlock(nn.Module):
    """
    Residual Block module for a vision model.

    Args:
        in_channels (int): Number of input channels.
        out_channels (int): Number of output channels.
        stride (int, optional): Stride value for the convolutional layers. Defaults to 1.
    """

    def __init__(self, in_channels, out_channels, stride=1):
        super().__init__()
        self.conv1 = nn.Conv2d(
            in_channels, out_channels, kernel_size=3, stride=stride, padding=1
        )
        self.bn1 = nn.BatchNorm2d(out_channels)
        self.conv2 = nn.Conv2d(
            out_channels, out_channels, kernel_size=3, padding=1
        )
        self.bn2 = nn.BatchNorm2d(out_channels)

        self.shortcut = nn.Sequential()
        if stride != 1 or in_channels != out_channels:
            self.shortcut = nn.Sequential(
                nn.Conv2d(
                    in_channels, out_channels, kernel_size=1, stride=stride
                ),
                nn.BatchNorm2d(out_channels),
            )

    def forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = self.bn2(self.conv2(out))
        out += self.shortcut(x)
        out = F.relu(out)
        return out


class VisionRewardModel(nn.Module):
    """
    VisionRewardModel is a neural network model that extracts image features and predicts rewards.

    Args:
        None

    Attributes:
        layer1 (ResidualBlock): The first residual block for image feature extraction.
        layer2 (ResidualBlock): The second residual block for image feature extraction.
        layer3 (ResidualBlock): The third residual block for image feature extraction.
        layer4 (ResidualBlock): The fourth residual block for image feature extraction.
        fc1 (nn.Linear): The fully connected layer for feature transformation.
        fc2 (nn.Linear): The fully connected layer for reward prediction.

    Methods:
        forward(x): Performs forward pass through the network.

    """

    def __init__(self):
        super().__init__()

        # Image Feature Extractor
        self.layer1 = ResidualBlock(3, 64)
        self.layer2 = ResidualBlock(64, 128, 2)
        self.layer3 = ResidualBlock(128, 256, 2)
        self.layer4 = ResidualBlock(256, 512, 2)
        self.fc1 = nn.Linear(512, 128)

        # Reward predictor
        self.fc2 = nn.Linear(128, 1)

    def forward(self, x):
        out = self.layer1(x)
        out = self.layer2(out)
        out = self.layer3(out)
        out = self.layer4(out)
        out = F.avg_pool2d(out, 4)
        out = out.view(out.size(0), -1)
        out = F.relu(self.fc1(out))
        reward = self.fc2(out)
        return reward


# Example usage

# # 1. Example for ResidualBlock
# res_block = ResidualBlock(in_channels=3, out_channels=64)
# sample_tensor = torch.randn(8, 3, 32, 32)
# output_tensor = res_block(sample_tensor)

# # 2. Example for VisionRewardModel
# vision_reward_model = VisionRewardModel()
# sample_image = torch.randn(8, 3, 32, 32)
# predicted_rewards = vision_reward_model(sample_image)

# print(output_tensor.shape, predicted_rewards.shape)



from torch import Tensor, nn


def add_norm(x, dim: int, residual: Tensor):
    """_summary_

    Args:
        x (_type_): _description_
        dim (int): _description_
        residual (Tensor): _description_

    Returns:
        _type_: _description_


    Example:
    x = torch.randn(1, 10, 10)
    y = torch.randn(1, 10, 10)
    model = add_norm(x, 10, y)
    print(model)
    """
    layer = nn.Sequential(nn.LayerNorm(dim))
    return layer(x) + residual



import torch
import torch.nn as nn
from einops import repeat


class TextVideoAttentionFusion(nn.Module):
    """
    Text-Video Attention Fusion

    Args:
        text_features (int): Text features
        video_features (int): Video features

    Shape:
        - text: :math:`(B, S_T, T_F)`
        - video: :math:`(B, S_V, H*W, V_F)`
        - Output: :math:`(B, S_T, S_V, V_F)`

    Examples::
        >>> from zeta.nn.modules import TextVideoAttentionFusion
        >>> import torch
        >>> x = torch.rand(1, 10, 512)
        >>> y = torch.rand(1, 20, 196, 512)
        >>> model = TextVideoAttentionFusion(512, 512)
        >>> out = model(x, y)
        >>> print(out.shape)
        torch.Size([1, 10, 20, 512])

    """

    def __init__(self, text_features, video_features):
        super().__init__()

        # A linear layer for calculating attention scores
        self.linear = nn.Linear(text_features + video_features, 1)

    def forward(self, text, video):
        """forward method"""
        # text: [batch_size, seq_len_text, text_features]
        # video: [batch_size, seq_len_video, h*w, video_features]

        # Get the dimensions
        batch_size, seq_len_video, hw, video_features = video.shape
        _, seq_len_text, text_features = text.shape

        # Using einops to repeat the tensors for matching dimensions
        text_expanded = repeat(
            text, "b st tf -> b st sv hw tf", sv=seq_len_video, hw=hw
        )
        video_expanded = repeat(
            video, "b sv hw vf -> b st sv hw vf", st=seq_len_text
        )

        # Concatenating expanded text tensor and video tensor
        concat_features = torch.cat(
            [text_expanded, video_expanded], dim=-1
        )  # [batch_size, seq_len_text, seq_len_video, h*w, text_features + video_features]

        # Computing attention scores and weights
        attention_scores = self.linear(concat_features).squeeze(
            -1
        )  # [batch_size, seq_len_text, seq_len_video, h*w]
        attention_weights = (
            torch.nn.functional.softmax(attention_scores.flatten(2), dim=-1)
            .view(batch_size, seq_len_text, seq_len_video, hw)
            .unsqueeze(-1)
        )  # [batch_size, seq_len_text, seq_len_video, h*w, 1]

        # Using einsum for the weighted sum across sequence lengths and spatial dimensions
        fused = torch.einsum(
            "btvhj,btvhj->btvj", video_expanded, attention_weights
        )  # [batch_size, seq_len_text, seq_len_video, video_features]

        return fused


# x = torch.rand(1, 10, 512)
# y = torch.rand(1, 20, 196, 512)
# model = TextVideoAttentionFusion(512, 512)
# out = model(x, y)
# print(out)



from torch import nn


def SimpleFeedForward(dim: int, hidden_dim: int, dropout=0.1):
    """
    Feedforward neural network with LayerNorms and GELU activations


    Flow:
    layer_norm -> linear -> gelu -> linear -> dropout


    Args:
        dim (int): Input dimension
        hidden_dim (int): Hidden dimension
        dropout (float): Dropout probability

    Usage:
    >>> model = SimpleFeedForward(768, 2048, 0.1)
    >>> x = torch.randn(1, 768)
    >>> model(x).shape
    torch.Size([1, 768])
    """
    return nn.Sequential(
        nn.LayerNorm(dim),
        nn.Linear(dim, hidden_dim),
        nn.GELU(),
        nn.Linear(hidden_dim, dim),
        nn.Dropout(dropout),
    )



import torch
from einops import rearrange, repeat
from einops.layers.torch import Rearrange
from torch import nn

# helpers


def pair(t):
    return t if isinstance(t, tuple) else (t, t)


# classes


class LayerNorm(nn.Module):
    def __init__(
        self,
        dim: int,
        eps: float = 1e-5,
        elementwise_affine: bool = True,
    ):
        super().__init__()
        self.layer_norm = nn.LayerNorm(dim, eps, elementwise_affine)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.layer_norm(x)


class FeedForward(nn.Module):
    def __init__(self, dim, hidden_dim, dropout=0.0):
        super().__init__()
        self.net = nn.Sequential(
            nn.LayerNorm(dim),
            nn.Linear(dim, hidden_dim),
            nn.GELU(),
            nn.Dropout(dropout),
            nn.Linear(hidden_dim, dim),
            nn.Dropout(dropout),
        )

    def forward(self, x):
        return self.net(x)


class Attention(nn.Module):
    def __init__(self, dim, heads=8, dim_head=64, dropout=0.0):
        super().__init__()
        inner_dim = dim_head * heads
        project_out = not (heads == 1 and dim_head == dim)

        self.heads = heads
        self.scale = dim_head**-0.5

        self.norm = nn.LayerNorm(dim)
        self.norm_k = nn.LayerNorm(dim_head)
        self.norm_v = nn.LayerNorm(dim_head)

        self.attend = nn.Softmax(dim=-1)
        self.dropout = nn.Dropout(dropout)

        self.to_qkv = nn.Linear(dim, inner_dim * 3, bias=False)

        self.to_out = (
            nn.Sequential(nn.Linear(inner_dim, dim), nn.Dropout(dropout))
            if project_out
            else nn.Identity()
        )

    def forward(self, x):
        x = self.norm(x)

        qkv = self.to_qkv(x).chunk(3, dim=-1)
        q, k, v = map(
            lambda t: rearrange(t, "b n (h d) -> b h n d", h=self.heads), qkv
        )

        # #normalize key and values,  QK Normalization
        k = self.norm_k(k)
        v = self.norm_v(v)

        dots = torch.matmul(q, k.transpose(-1, -2)) * self.scale

        attn = self.attend(dots)
        attn = self.dropout(attn)

        out = torch.matmul(attn, v)
        out = rearrange(out, "b h n d -> b n (h d)")
        return self.to_out(out)


class Transformer(nn.Module):
    def __init__(self, dim, depth, heads, dim_head, mlp_dim, dropout=0.0):
        super().__init__()
        self.norm = nn.LayerNorm(dim)
        self.layers = nn.ModuleList([])
        for _ in range(depth):
            self.layers.append(
                nn.ModuleList(
                    [
                        Attention(
                            dim, heads=heads, dim_head=dim_head, dropout=dropout
                        ),
                        FeedForward(dim, mlp_dim, dropout=dropout),
                    ]
                )
            )

    def forward(self, x):
        for attn, ff in self.layers:
            # layernorm before attention
            x = self.norm(x)

            # parallel
            x = x + attn(x) + ff(x)

        return self.norm(x)


class MegaVit(nn.Module):
    """
    MegaVit model from https://arxiv.org/abs/2106.14759

    Args:
    -------
    image_size: int
        Size of image
    patch_size: int
        Size of patch
    num_classes: int
        Number of classes
    dim: int
        Dimension of embedding
    depth: int
        Depth of transformer
    heads: int
        Number of heads
    mlp_dim: int
        Dimension of MLP
    pool: str
        Type of pooling
    channels: int
        Number of channels
    dim_head: int
        Dimension of head
    dropout: float
        Dropout rate
    emb_dropout: float
        Dropout rate for embedding

    Returns:
    --------
    torch.Tensor
        Predictions

    Methods:
    --------
    forward(img: torch.Tensor) -> torch.Tensor:
        Forward pass

    Architecture:
    -------------
    1. Input image is passed through a patch embedding layer
    2. Positional embedding is added
    3. Dropout is applied
    4. Transformer is applied
    5. Pooling is applied
    6. MLP head is applied
    7. Output is returned

    Usage
    -----
    >>> model = MegaVit(
    ...     image_size = 256,
    ...     patch_size = 32,
    ...     num_classes = 1000,
    ...     dim = 512,
    ...     depth = 6,
    ...     heads = 8,
    ...     mlp_dim = 1024,
    ...     dropout = 0.1,
    ...     emb_dropout = 0.1
    ... )
    >>> img = torch.randn(1, 3, 256, 256)
    >>> preds = model(img) # (1, 1000)

    References:
    -----------
    [1] https://arxiv.org/abs/2106.14759

    """

    def __init__(
        self,
        *,
        image_size,
        patch_size,
        num_classes,
        dim,
        depth,
        heads,
        mlp_dim,
        pool="cls",
        channels=3,
        dim_head=64,
        dropout=0.0,
        emb_dropout=0.0,
    ):
        super().__init__()
        image_height, image_width = pair(image_size)
        patch_height, patch_width = pair(patch_size)

        assert (
            image_height % patch_height == 0 and image_width % patch_width == 0
        ), "Image dimensions must be divisible by the patch size."

        num_patches = (image_height // patch_height) * (
            image_width // patch_width
        )
        patch_dim = channels * patch_height * patch_width
        assert pool in {
            "cls",
            "mean",
        }, "pool type must be either cls (cls token) or mean (mean pooling)"

        self.to_patch_embedding = nn.Sequential(
            Rearrange(
                "b c (h p1) (w p2) -> b (h w) (p1 p2 c)",
                p1=patch_height,
                p2=patch_width,
            ),
            nn.LayerNorm(patch_dim),
            nn.Linear(patch_dim, dim),
            nn.LayerNorm(dim),
        )

        self.pos_embedding = nn.Parameter(torch.randn(1, num_patches + 1, dim))
        self.cls_token = nn.Parameter(torch.randn(1, 1, dim))
        self.dropout = nn.Dropout(emb_dropout)

        self.transformer = Transformer(
            dim, depth, heads, dim_head, mlp_dim, dropout
        )

        self.pool = pool
        self.to_latent = nn.Identity()

        self.mlp_head = nn.Linear(dim, num_classes)

    def forward(self, img):
        x = self.to_patch_embedding(img)
        b, n, _ = x.shape

        cls_tokens = repeat(self.cls_token, "1 1 d -> b 1 d", b=b)
        x = torch.cat((cls_tokens, x), dim=1)
        x += self.pos_embedding[:, : (n + 1)]
        x = self.dropout(x)

        x = self.transformer(x)

        x = x.mean(dim=1) if self.pool == "mean" else x[:, 0]

        x = self.to_latent(x)
        return self.mlp_head(x)



import torch
from torch import nn


class SkipConnection(nn.Module):
    """
    A helper class for implementing skip connections.
    """

    def __init__(self):
        super().__init__()

    def forward(self, x1, x2):
        return x1 + x2


class MultiModalAdapterDenseNetwork(nn.Module):
    """
    Multi-modal adapter dense network that takes a tensor of shape (batch_size, dim) and returns a tensor of shape (batch_size, dim).

    Flow:
    x -> norm -> linear 1 -> silu -> concate -> linear 2 -> skip connection -> output

    Args:
        dim (int): The input dimension.
        hidden_dim (int): The hidden dimension.
        depth (int): The depth of the network.
        activation (nn.Module): The activation function.

    Methods:
        forward(x: torch.Tensor) -> torch.Tensor: The forward pass of the network.

    Example:
        >>> from zeta.nn import MultiModalAdapterDenseNetwork
        >>> mm_adapter = MultiModalAdapterDenseNetwork(
        ...     dim=512,
        ...     hidden_dim=1024,
        ...     depth=3,
        ... )
        >>> output = mm_adapter(x)
        >>> print(output.shape)
        torch.Size([1, 1024, 512])


    """

    def __init__(
        self,
        dim: int = None,
        hidden_dim: int = None,
        depth: int = None,
        activation: nn.Module = nn.SiLU(),
    ):
        super().__init__()
        self.dim = dim
        self.hidden_dim = hidden_dim
        self.out_dim = dim
        self.depth = depth
        self.activation = activation

        self.layers = nn.ModuleList([])
        self.norm = nn.LayerNorm(self.dim)
        self.proj = nn.Linear(self.dim, self.dim)

        # Define layers
        self.layers = nn.ModuleList([])
        for _ in range(depth):
            self.layers.append(
                nn.Sequential(
                    nn.LayerNorm(self.dim),
                    nn.Linear(self.dim, self.hidden_dim),
                    nn.SiLU(),
                    nn.Linear(self.hidden_dim, dim),
                )
            )
        self.skip_connections = SkipConnection()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the network.
        """
        for layer in self.layers:
            # Apply dense layer block ops
            y = layer(x)

            # Add the input of the block to it's output(skip connection)
            x = self.skip_connections(x, y)
        return x



"""
From https://github.com/milesial/Pytorch-UNet/blob/master/unet/unet_parts.py

"""

import torch
import torch.nn.functional as F
from torch import nn


class DoubleConv(nn.Module):
    def __init__(self, in_channels, out_channels, mid_channels=None):
        super().__init__()
        if not mid_channels:
            mid_channels = out_channels
        self.double_conv = nn.Sequential(
            nn.Conv2d(
                in_channels, mid_channels, kernel_size=3, padding=1, bias=False
            ),
            nn.BatchNorm2d(mid_channels),
            nn.ReLU(inplace=True),
            nn.Conv2d(
                mid_channels, out_channels, kernel_size=3, padding=1, bias=False
            ),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True),
        )

    def forward(self, x):
        return self.double_conv(x)


class Down(nn.Module):
    def __init__(self, in_channels, out_channels):
        super().__init__()
        self.maxpool_conv = nn.Sequential(
            nn.MaxPool2d(2), DoubleConv(in_channels, out_channels)
        )

    def forward(self, x):
        return self.maxpool_conv(x)


class Up(nn.Module):
    def __init__(self, in_channels, out_channels, bilinear=True):
        super().__init__()

        if bilinear:
            self.up = nn.Upsample(
                scale_factor=2, mode="bilinear", align_corners=True
            )
            self.conv = DoubleConv(in_channels, out_channels, in_channels // 2)
        else:
            self.up = nn.ConvTranspose2d(
                in_channels, in_channels // 2, kernel_size=2, stride=2
            )
            self.conv = DoubleConv(in_channels, out_channels)

    def forward(self, x1, x2):
        x1 = self.up(x1)

        diffy = x2.size()[2] - x1.size()[2]
        diffx = x2.size()[3] - x1.size()[3]

        x1 = F.pad(
            x1, [diffx // 2, diffx - diffx // 2, diffy // 2, diffy - diffy // 2]
        )
        x = torch.cat([x2, x1], dim=1)
        return self.conv(x)


class OutConv(nn.Module):
    def __init__(self, in_channels, out_channels):
        super().__init__()
        self.conv = nn.Conv2d(in_channels, out_channels, kernel_size=1)

    def forward(self, x):
        return self.conv(x)


class Unet(nn.Module):
    """
    UNET model

    Flow:
        1. Downsample
        2. Upsample
        3. Output

    Args:
        n_channels (int): Number of input channels
        n_classes (int): Number of output channels
        bilinear (bool): If True, use bilinear interpolation for upsampling

    Methods:
        forward: Forward pass
        use_checkpointing: Use checkpointing to save memory

    Examples:
        >>> import torch
        >>> from zeta.nn.modules.unet import Unet
        >>> model = Unet(1, 2)
        >>> x = torch.randn(1, 1, 572, 572)
        >>> y = model(x)
        >>> y.shape
        torch.Size([1, 2, 388, 388])


    """

    def __init__(self, n_channels, n_classes, bilinear=False):
        super().__init__()
        self.n_channels = n_channels
        self.n_classes = n_classes
        self.bilinear = bilinear

        self.inc = DoubleConv(n_channels, 64)
        self.down1 = Down(64, 128)
        self.down2 = Down(128, 256)
        self.down3 = Down(256, 512)
        factor = 2 if bilinear else 1
        self.down4 = Down(512, 1024 // factor)

        self.up1 = Up(1024, 512 // factor, bilinear)
        self.up2 = Up(512, 256 // factor, bilinear)
        self.up3 = Up(256, 128 // factor, bilinear)
        self.up4 = Up(128, 64, bilinear)
        self.outc = OutConv(64, n_classes)

    def forward(self, x):
        """
        Forward pass

        Args:
            x (torch.Tensor): Input tensor


        Returns:
            torch.Tensor: Output tensor



        """
        x1 = self.inc(x)
        x2 = self.down1(x1)
        x3 = self.down2(x2)
        x4 = self.down3(x3)
        x5 = self.down4(x4)
        x = self.up1(x5, x4)
        x = self.up2(x, x3)
        x = self.up3(x, x2)
        x = self.up4(x, x1)
        logits = self.outc(x)
        return logits

    def use_checkpointing(self):
        """
        Use checkpointing to save memory


        """
        self.inc = torch.utils.checkpoint(self.inc)
        self.down1 = torch.utils.checkpoint(self.down1)
        self.down2 = torch.utils.checkpoint(self.down2)
        self.down3 = torch.utils.checkpoint(self.down3)
        self.down4 = torch.utils.checkpoint(self.down4)

        self.up1 = torch.utils.checkpoint(self.up1)
        self.up2 = torch.utils.checkpoint(self.up2)
        self.up3 = torch.utils.checkpoint(self.up3)
        self.up4 = torch.utils.checkpoint(self.up4)
        self.outc = torch.utils.checkpoint(self.outc)



from torch import Tensor, nn


class MonarchMLP(nn.Module):
    """
    A sparse MLP from this paper: https://hazyresearch.stanford.edu/blog/2024-01-11-m2-bert-retrieval

    Example:
        >>> x = torch.randn(1, 3, 32, 32)
        >>> model = MonarchMLP()
        >>> out = model(x)
        >>> print(out)
    """

    def __init__(
        self,
    ):
        super().__init__()

        self.glu = nn.GLU()
        self.gelu = nn.GELU()

    def forward(self, x: Tensor):
        """
        Forward pass of the MonarchMLP model.

        Args:
            x (Tensor): Input tensor.

        Returns:
            Tensor: Output tensor after passing through GLU and GELU activation functions.
        """
        x = self.glu(x)
        x = self.gelu(x)
        return x



# replace some of the activation functions from sigmoid to exponential function - e ^ x
# Memory saving: make the memory larger --> associate memory --> increase


from torch import nn, Tensor


class SimpleRNN(nn.Module):
    """
    A simple recurrent neural network module.

    Args:
        dim (int): The input dimension.
        hidden_dim (int): The dimension of the hidden state.
    """

    def __init__(
        self,
        dim: int = None,
        hidden_dim: int = None,
    ):
        super().__init__()
        self.dim = dim
        self.hidden_dim = hidden_dim

        self.act = nn.Tanh()

    def forward(self, x: Tensor) -> Tensor:
        """
        Forward pass of the simple RNN module.

        Args:
            x (Tensor): The input tensor of shape (batch_size, sequence_length, input_dim).

        Returns:
            Tensor: The output tensor of shape (batch_size, sequence_length, hidden_dim).
        """
        b, s, d = x.shape

        h = self.act(x)

        return h



import torch
from torch import nn, Tensor


class CoPE(nn.Module):
    def __init__(self, npos_max: int, dim: int = None):
        super().__init__()
        self.npos_max = npos_max
        self.pos_emb = nn.parameter.Parameter(torch.zeros(1, dim, npos_max))

    def forward(self, query: Tensor, attn_logits: Tensor) -> Tensor:
        # compute positions
        gates = torch.sigmoid(attn_logits)
        pos = gates.flip(-1).cumsum(dim=-1).flip(-1)
        pos = pos.clamp(max=self.npos_max - 1)
        # interpolate from integer positions
        pos_ceil = pos.ceil().long()
        pos_floor = pos.floor().long()
        logits_int = torch.matmul(query, self.pos_emb)
        logits_ceil = logits_int.gather(-1, pos_ceil)
        logits_floor = logits_int.gather(-1, pos_floor)
        w = pos - pos_floor
        return logits_ceil * w + logits_floor * (1 - w)


# x = torch.randn(1, 5, 10)
# attn_logits = torch.randn(1, 5, 10)

# cope = CoPE(5, 10)
# out = cope(x, attn_logits)
# print(out)



import math
from typing import Optional

import torch
from torch import nn
from torch.nn.parameter import Parameter


class CombinedLinear(nn.Module):
    r"""Applies a linear transformation to the incoming data: :math:`y = xA^T + b`

    Compared to torch.nn.Linear, uses a combined parameter for both the weight and bias terms.
    This is useful for reducing the number of parameters for methods that exploit tensor structure, such as Shampoo.

    This module supports :ref:`TensorFloat32<tf32_on_ampere>`.
    On certain ROCm devices, when using float16 inputs this module will use :ref:`different precision<fp16_on_mi200>` for backward.

    Args:
        in_features (int): size of each input sample
        out_features (int): size of each output sample
        bias (bool): If set to ``False``, the layer will not learn an additive bias.
            Default: ``True``

    Shape:
        - Input: :math:`(*, H_{in})` where :math:`*` means any number of
          dimensions including none and :math:`H_{in} = \text{in\_features}`.
        - Output: :math:`(*, H_{out})` where all but the last dimension
          are the same shape as the input and :math:`H_{out} = \text{out\_features}`.

    Attributes:
        combined_weight: the learnable weights of the module of shape
            :math:`(\text{out\_features}, \text{in\_features\_with\_bias})`.
            If :attr:`bias` is ``True``, :math:`\text{in\_features\_with\_bias} = \text{in\_features} + 1`;
            otherwise, :math:`\text{in\_features\_with\_bias} = \text{in\_features}`.
            The values are initialized from :math:`\mathcal{U}(-\sqrt{k}, \sqrt{k})`, where
            :math:`k = \frac{1}{\text{in\_features}}`
        bias: the flag for using the bias term

    Call Args:
        input (torch.Tensor): tensor of shape (B, I),
            where B is batch size and I is number of elements in each input sample (i.e. `in_features`).

    Returns:
        output (torch.Tensor): tensor of shape (B, O), where B is batch size
            and O is number of elements in each output sample (i.e. `out_features`).

    Examples::
        >>> from exa import CombinedLinear
        >>> m = CombinedLinear(20, 30)
        >>> input = torch.randn(128, 20)
        >>> output = m(input)
        >>> print(output.size())
        torch.Size([128, 30])
    """

    __constants__ = ["in_features", "out_features"]
    in_features: int
    out_features: int

    def __init__(
        self,
        in_features: int,
        out_features: int,
        bias: bool = True,
        device: Optional[torch.device] = None,
        dtype: Optional[torch.dtype] = None,
    ) -> None:
        factory_kwargs = {"device": device, "dtype": dtype}

        super().__init__()

        self.in_features = in_features
        self.out_features = out_features
        self.in_features_with_bias: int = (
            in_features + 1 if bias else in_features
        )
        self.bias = bias
        self.combined_weight = Parameter(
            torch.empty(
                (self.out_features, self.in_features_with_bias),
                **factory_kwargs,
            )
        )
        self.reset_parameters()

    def reset_parameters(self) -> None:
        if self.bias:
            torch.nn.init.kaiming_uniform_(
                self.combined_weight[:, :-1], a=math.sqrt(5)
            )
            fan_in, _ = torch.nn.init._calculate_fan_in_and_fan_out(
                self.combined_weight[:, :-1]
            )
            bound = 1 / math.sqrt(fan_in) if fan_in > 0 else 0
            torch.nn.init.uniform_(self.combined_weight[:, -1], -bound, bound)
        else:
            torch.nn.init.kaiming_uniform_(self.combined_weight, a=math.sqrt(5))

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        if self.bias:
            return torch.nn.functional.linear(
                input, self.combined_weight[:, :-1], self.combined_weight[:, -1]
            )
        else:
            return torch.nn.functional.linaer(input, self.combined_weight, None)

    def extra_repr(self) -> str:
        return (
            "in_features={}, out_features={}, in_features_with_bias={}".format(
                self.in_features, self.out_features, self.in_features_with_bias
            )
        )



import torch
import torch.nn as nn


class FlexiConv(nn.Module):
    """
    FlexiConv is an experimental and flexible convolutional layer that adapts to the input data.

    This layer uses parameterized Gaussian functions to weigh the importance of each pixel
    in the receptive field and applies a depthwise separable convolution for efficiency.

    Args:
        in_channels (int): Number of channels in the input image.
        out_channels (int): Number of channels produced by the convolution.
        kernel_size (int or tuple): Size of the convolving kernel.
        stride (int or tuple, optional): Stride of the convolution. Default: 1
        padding (int or tuple, optional): Zero-padding added to both sides of the input. Default: 0

    Example:
        >>> flexiconv = FlexiConv(in_channels=3, out_channels=64, kernel_size=3, stride=1, padding=1)
        >>> input_tensor = torch.randn(1, 3, 224, 224)  # Example input batch
        >>> output = flexiconv(input_tensor)
        >>> output.shape
    """

    def __init__(
        self, in_channels, out_channels, kernel_size, stride=1, padding=0
    ):
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = (
            kernel_size
            if isinstance(kernel_size, tuple)
            else (kernel_size, kernel_size)
        )
        self.stride = stride
        self.padding = padding

        # Gaussian weights
        self.gaussian_weights = nn.Parameter(
            torch.randn(in_channels, *self.kernel_size)
        )

        # Depthwise separable convolution
        self.depthwise = nn.Conv2d(
            in_channels,
            in_channels,
            kernel_size=self.kernel_size,
            stride=self.stride,
            padding=self.padding,
            groups=in_channels,
        )
        self.pointwise = nn.Conv2d(
            in_channels, out_channels, kernel_size=1, stride=1
        )

        # Initialization of the parameters
        self._reset_parameters()

    def _reset_parameters(self):
        nn.init.kaiming_normal_(
            self.depthwise.weight, mode="fan_out", nonlinearity="relu"
        )
        nn.init.constant_(self.depthwise.bias, 0)
        nn.init.kaiming_normal_(
            self.pointwise.weight, mode="fan_out", nonlinearity="relu"
        )
        nn.init.constant_(self.pointwise.bias, 0)
        nn.init.normal_(self.gaussian_weights, mean=0, std=0.1)

    def forward(self, x):
        """
        Forward pass of the FlexiConv layer.

        Args:
            x (torch.Tensor): The input tensor.

        Returns:
            torch.Tensor: The result of the flexible convolution.
        """
        # Apply depthwise convolution
        depthwise_out = self.depthwise(x)

        # Generate a Gaussian mask for each channel
        gaussian_mask = torch.exp(-torch.square(self.gaussian_weights))

        # Use einsum to apply the gaussian mask with depthwise convolution output.
        # 'bcij,ckl->bcijkl' denotes a mapping from the batch and channel dimensions (bc),
        # input spatial dimensions (ij), and the kernel dimensions (kl) to a combined output tensor.
        combined = torch.einsum(
            "bcij,ckl->bcijkl", depthwise_out, gaussian_mask
        )

        # Sum over the kernel dimensions to apply the gaussian mask
        weighted_out = combined.sum(dim=-2).sum(dim=-2)

        # Apply pointwise convolution
        out = self.pointwise(weighted_out)

        return out



from typing import List

from torch import Tensor, nn

from zeta.nn.modules.ws_conv2d import WSConv2d


class NFNStem(nn.Module):
    """
    NFNStem module represents the stem of the NFN (Neural Filter Network) architecture.

    Args:
        in_channels (List[int]): List of input channel sizes for each layer. Default is [3, 16, 32, 64].
        out_channels (List[int]): List of output channel sizes for each layer. Default is [16, 32, 64, 128].
        kernel_size (int): Size of the convolutional kernel. Default is 3.
        stride (List[int]): List of stride values for each layer. Default is [2, 1, 1, 2].
        activation (nn.Module): Activation function to be applied after each convolutional layer. Default is nn.GELU().

    Examples:
        >>> x = torch.randn(1, 3, 224, 224)
        >>> model = NFNStem()
        >>> out = model(x)
        >>> print(out.shape)
        torch.Size([1, 128, 28, 28])
    """

    def __init__(
        self,
        in_channels: List[int] = [3, 16, 32, 64],
        out_channels: List[int] = [16, 32, 64, 128],
        kernel_size: int = 3,
        stride: List[int] = [2, 1, 1, 2],
        activation: nn.Module = nn.GELU(),
    ):
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.activation = activation
        self.kernel_size = kernel_size
        self.stride = stride

        self.conv0 = WSConv2d(
            in_channels=self.in_channels[0],
            out_channels=self.out_channels[0],
            kernel_size=3,
            stride=self.stride[0],
        )
        self.conv1 = WSConv2d(
            in_channels=self.in_channels[1],
            out_channels=self.out_channels[1],
            kernel_size=kernel_size,
            stride=self.stride[1],
        )
        self.conv2 = WSConv2d(
            in_channels=self.in_channels[2],
            out_channels=self.out_channels[2],
            kernel_size=kernel_size,
            stride=self.stride[2],
        )
        self.conv3 = WSConv2d(
            in_channels=self.in_channels[3],
            out_channels=out_channels[3],
            kernel_size=kernel_size,
            stride=self.stride[3],
        )

    def forward(self, x: Tensor):
        """Forward pass of the NFNStem module.

        Args:
            x (Tensor): _description_

        Returns:
            _type_: _description_
        """
        out = self.activation(self.conv0(x))
        out = self.activation(self.conv1(out))
        out = self.activation(self.conv2(out))
        out = self.conv3(out)
        return out



from typing import List, TypeVar

import torch
from einops import rearrange

Tensor = TypeVar("Tensor", bound=torch.Tensor)


class Tensor(torch.nn.Module):
    def __init__(
        self,
        data: torch.Tensor,
        shape: List[str],
        to: List[str],
    ):
        super().__init__()
        self.data = data
        self.shape = shape
        self.to = to

    def __call__(self):
        shape = " ".join(self.shape)
        to = "".join(self.to)

        return rearrange(
            self.data,
            shape + " -> " + to,
        )


# # Example
# x = torch.randn(2, 4, 6, 8)

# model = Tensor(
#     data=x,
#     shape=["b d s h"],
#     to=['b h s d']
# )

# out = model()
# print(out)



import torch
from einops import rearrange
import torch.nn.functional as F
from torch import Tensor
from torch import nn
from zeta.structs.auto_regressive_wrapper import AutoRegressiveWrapper
from typing import List
from einops import reduce


def exists(val):
    return val is not None


def return_loss_text(
    x: Tensor, logits: Tensor, labels: Tensor, ignore_index, mask: Tensor
):
    """
    Computes the cross-entropy loss between the predicted logits and the target labels.

    Args:
        logits (Tensor): The predicted logits of shape (batch_size, num_classes, sequence_length).
        labels (Tensor): The target labels of shape (batch_size, sequence_length).
        ignore_index (int): The index to ignore when computing the loss.

    Returns:
        Tensor: The computed cross-entropy loss.
    """
    _seq, labels = x[:, :-1], x[:, 1:]

    labels = labels.masked_fill(~mask[:, 1:], ignore_index)

    loss = F.cross_entropy(
        rearrange(logits, "b n c -> b c n"), labels, ignore_index=ignore_index
    )

    return loss


def add_masking_llm(x: Tensor, mask: Tensor, ignore_index: int):
    """
    Adds masking to the input tensor.

    Args:
        x (Tensor): The input tensor.
        ignore_index (int): The index to ignore.

    Returns:
        Tensor: The masked input tensor.
    """
    ...


def calc_z_loss(
    pre_softmax_attns: List[Tensor], mask: Tensor = None, weight: float = 1.0
):
    lse = 0.0

    for attn in pre_softmax_attns:
        lse = lse + attn.logsumexp(dim=-1)

    loss = torch.square(lse)
    loss = reduce(loss, "b h n -> b n", "sum")

    if not exists(mask):
        return loss.mean() * weight

    loss = loss[mask].sum() / mask.sum().clamp(min=1e-5)
    return loss * weight


def max_neg_value(tensor: Tensor):
    return -torch.finfo(tensor.dtype).max


def l2norm(x: Tensor, groups: int = 1):
    """
    Applies L2 normalization to the input tensor.

    Args:
        x (Tensor): The input tensor to be normalized.
        groups (int, optional): The number of groups to divide the input tensor into. Defaults to 1.

    Returns:
        Tensor: The normalized tensor.

    """
    x = rearrange(x, "... (g d) -> ... g d", g=groups)
    x = F.normalize(x, p=2, dim=-1)
    return rearrange(x, "... g d -> ... (g d)")


class TextTokenEmbedding(nn.Module):
    def __init__(
        self,
        dim: int,
        num_tokens: int,
        l2norm_embed: bool = True,
    ):
        """
        Initializes a TextTokenEmbedding module.

        Args:
            dim (int): The dimension of the embedding.
            num_tokens (int): The number of tokens in the vocabulary.
            l2norm_embed (bool, optional): Whether to apply L2 normalization to the embeddings. Defaults to True.
        """
        super().__init__()
        self.dim = dim
        self.num_tokens = num_tokens
        self.l2norm_embed = l2norm_embed
        self.embed = nn.Embedding(num_tokens, dim)

    def forward(self, x: Tensor):
        """
        Forward pass of the TextTokenEmbedding module.

        Args:
            x (Tensor): The input tensor of shape (batch_size, sequence_length).

        Returns:
            Tensor: The embedded tensor of shape (batch_size, sequence_length, dim).
        """
        token_embed = self.embed(x.long())
        return l2norm(token_embed) if self.l2norm_embed else token_embed


def dropout_seq(seq: Tensor, mask: Tensor, dropout: float = 0.0):
    """
    Applies dropout to a sequence of tensors.

    Args:
        seq (Tensor): The input sequence tensor of shape (batch_size, sequence_length, ...).
        mask (Tensor): The mask tensor of shape (batch_size, sequence_length) indicating which elements to keep.
        dropout (float, optional): The dropout probability. Defaults to 0.

    Returns:
        Tuple[Tensor, Tensor]: A tuple containing the modified sequence tensor and the modified mask tensor.

    """
    b, n, *_, device = *seq.shape, seq.device
    logits = torch.randn(b, n, device=device)

    if exists(mask):
        mask_value = max_neg_value(logits)
        logits = logits.masked_fill(~mask, mask_value)

    keep_prob = 1.0 - dropout
    num_keep = max(1, int(keep_prob * n))
    keep_indices = logits.topk(num_keep, dim=1).indices

    batch_indices = torch.arange(b, device=device)
    batch_indices = rearrange(batch_indices, "b -> b 1")

    seq = seq[batch_indices, keep_indices]

    if exists(mask):
        seq_counts = mask.sum(dim=-1)
        seq_keep_counts = torch.ceil(seq_counts * keep_prob).int()
        keep_mask = torch.arange(num_keep, device=device) < rearrange(
            seq_keep_counts, "b -> b 1"
        )

        mask = mask[batch_indices, keep_indices] & keep_mask

    return seq, mask


@torch.no_grad()
def transformer_generate(
    model: nn.Module,
    prompt: Tensor,
    temperature: float = 0.5,
    filter_threshold: float = 0.9,
    *args,
    **kwargs,
):
    """
    Generates text given a prompt.

    Args:
        model (nn.Module): The model to generate text.
        prompt (Tensor): The prompt tensor.

    Returns:
        Tensor: The generated text.
    """
    model = AutoRegressiveWrapper(net=model)

    return model.generate(
        prompt,
        filter_thres=filter_threshold,
        temperature=temperature,
        *args,
        **kwargs,
    )



from torch import nn


class SqueezeExcitation(nn.Module):
    """
    Squeeze-and-Excitation block.

    Parameters
    ---------
    in_planes : int
        the number of input channels
    reduced_dim : int
        the number of channels after the first convolution

    Attributes
    ----------
    se : nn.Sequential
        the sequential layers of the Squeeze-and-Excitation block

    Methods
    -------
    forward(x)

    Example:
    --------
    >>> x = torch.randn(1, 3, 256, 256)
    >>> model = SqueezeExcitation(3, 1)
    >>> output = model(x)
    >>> print(output.shape)



    """

    def __init__(self, in_planes, reduced_dim):
        super().__init__()
        self.se = nn.Sequential(
            nn.AdaptiveAvgPool2d(1),
            nn.Conv2d(in_planes, reduced_dim, 1),
            nn.ReLU6(inplace=True),
            nn.Conv2d(reduced_dim, in_planes, 1),
            nn.Sigmoid(),
        )

    def forward(self, x):
        """Forward pass for the Squeeze-and-Excitation block."""
        return x * self.se(x)



import torch
from torch import nn


class FusedDenseGELUDense(nn.Module):
    """FuseFusedDenseGELUDense

    Args
        dim (int): Input dimension
        dim_out (int): Output dimension
        bias (bool, optional): Bias. Defaults to True.
        has_fp16_weights (bool, optional): Use fp16 weights. Defaults to False.
        threshold (float, optional): Threshold for quantization. Defaults to 6.0.

    Examples:
        >>> x = torch.randn(1, 512)
        >>> model = FusedDenseGELUDense(512, 1024)
        >>> out = model(x)
        >>> out.shape
        torch.Size([1, 512])
    """

    def __init__(
        self,
        dim: int,
        dim_out: int,
        bias: bool = True,
        has_fp16_weights: bool = False,
        threshold: float = 6.0,
        *args,
        **kwargs,
    ):
        super().__init__()
        self.dim = dim
        self.dim_out = dim_out
        self.bias = bias
        self.has_fp16_weights = has_fp16_weights
        self.threshold = threshold

        try:
            import bitsandbytes as bnb

            # Using bitsandbytes for quantization
            self.dense1 = bnb.nn.Linear8bitLt(
                dim,
                dim_out,
                bias=bias,
                has_fp16_weights=has_fp16_weights,
                threshold=threshold,
                *args,
                **kwargs,
            )

            # Reverse
            self.dense2 = bnb.nn.Linear8bitLt(
                dim_out,
                dim,
                bias=bias,
                has_fp16_weights=has_fp16_weights,
                threshold=threshold,
                *args,
                **kwargs,
            )

        except ModuleNotFoundError:
            # Using torch.nn.Linear
            self.dense1 = nn.Linear(dim, dim_out, bias=bias * args, **kwargs)

            # Dense 2
            self.dense2 = nn.Linear(dim_out, dim, bias=bias * args, **kwargs)

        # Activation
        self.act = nn.GELU()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass

        Args:
            x (torch.Tensor): x input

        Returns:
            torch.Tensor: _description_
        """
        x = self.dense1(x)
        x = self.act(x)
        x = self.dense2(x)
        return x



from torch import nn, Tensor


class PixelShuffleDownscale(nn.Module):
    def __init__(self, downscale_factor: int = 2):
        """
        Initializes a PixelShuffleDownscale module.

        Args:
            downscale_factor (int): The factor by which the input will be downscaled.

        Example:
        >>> downscale_factor = 2
        >>> model = PixelShuffleDownscale(downscale_factor)
        >>> input_tensor = torch.rand(1, 256, 448, 448)
        >>> output_tensor = model(input_tensor)
        >>> print(output_tensor.shape)
        torch.Size([1, 64, 896, 896])
        """
        super(PixelShuffleDownscale, self).__init__()
        self.downscale_factor = downscale_factor
        # Initialize the pixel shuffle with an upscale factor which will actually be used to downscale
        self.pixel_shuffle = nn.PixelShuffle(upscale_factor=downscale_factor)

    def forward(self, x: Tensor) -> Tensor:
        """
        Performs a forward pass of the PixelShuffleDownscale module.

        Args:
            x (torch.Tensor): The input tensor with shape [batch_size, channels, height, width].

        Returns:
            torch.Tensor: The output tensor after downsampling using pixel shuffle.
        """
        # x should have a shape of [batch_size, channels, height, width]
        # We first need to adapt the number of channels so that pixel shuffle can be applied
        batch_size, channels, height, width = x.shape
        new_channels = channels // (self.downscale_factor**2)
        if new_channels * (self.downscale_factor**2) != channels:
            raise ValueError(
                "The number of channels must be divisible by"
                " (downscale_factor^2)"
            )

        # Reshape x to the shape expected by pixel shuffle
        x = x.reshape(
            batch_size, new_channels, self.downscale_factor**2, height, width
        )
        x = x.permute(0, 2, 1, 3, 4).contiguous()
        x = x.view(
            batch_size,
            new_channels * (self.downscale_factor**2),
            height,
            width,
        )

        # Apply pixel shuffle to reduce spatial dimensions and increase channel depth
        x = self.pixel_shuffle(x)

        return x


# # Example of usage
# downscale_factor = (
#     2  # This factor needs to be determined based on the required reduction
# )
# model = PixelShuffleDownscale(downscale_factor)
# input_tensor = torch.rand(1, 256, 448, 448)  # Example input tensor
# output_tensor = model(input_tensor)
# print(output_tensor.shape)  # This will print the shape of the output tensor



import torch
from torch import nn, Tensor


class SimpleLSTMCell(nn.Module):
    def __init__(self, dim: int, hidden_dim: int):
        """
        Simple LSTM cell implementation.

        Args:
            dim (int): The input dimension.
            hidden_dim (int): The hidden dimension.
        """
        super(SimpleLSTMCell, self).__init__()
        self.dim = dim
        self.hidden_dim = hidden_dim

        # Linear layers for input gate, forget gate, output gate, and cell state
        self.W_i = nn.Linear(dim, hidden_dim)
        self.U_i = nn.Linear(hidden_dim, hidden_dim)

        self.W_f = nn.Linear(dim, hidden_dim)
        self.U_f = nn.Linear(hidden_dim, hidden_dim)

        self.W_o = nn.Linear(dim, hidden_dim)
        self.U_o = nn.Linear(hidden_dim, hidden_dim)

        self.W_c = nn.Linear(dim, hidden_dim)
        self.U_c = nn.Linear(hidden_dim, hidden_dim)

    def forward(self, x: Tensor, h: Tensor, c: Tensor) -> Tensor:
        """
        Forward pass of the Simple LSTM cell.

        Args:
            x (Tensor): The input tensor of shape (batch_size, input_dim).
            h (Tensor): The previous hidden state tensor of shape (batch_size, hidden_dim).
            c (Tensor): The previous cell state tensor of shape (batch_size, hidden_dim).

        Returns:
            Tensor: The next hidden state tensor.
            Tensor: The next cell state tensor.
        """
        # Compute input gate
        i = torch.sigmoid(self.W_i(x) + self.U_i(h))

        # Compute forget gate
        f = torch.sigmoid(self.W_f(x) + self.U_f(h))

        # Compute output gate
        o = torch.sigmoid(self.W_o(x) + self.U_o(h))

        # Compute new cell candidate
        c_tilde = torch.tanh(self.W_c(x) + self.U_c(h))

        # Update cell state
        c_next = f * c + i * c_tilde

        # Update hidden state
        h_next = o * torch.tanh(c_next)

        return h_next, c_next


class SimpleLSTM(nn.Module):
    """
    Simple LSTM implementation.

    Args:
        dim (int): The input dimension.
        hidden_dim (int): The hidden dimension.
        depth (int): The number of LSTM layers.
        output_dim (int): The output dimension.
    """

    def __init__(self, dim: int, hidden_dim: int, depth: int, output_dim: int):
        super(SimpleLSTM, self).__init__()
        self.dim = dim
        self.hidden_dim = hidden_dim
        self.depth = depth

        # LSTM cells
        self.cells = nn.ModuleList(
            [
                SimpleLSTMCell(dim if i == 0 else hidden_dim, hidden_dim)
                for i in range(depth)
            ]
        )

        # Final output layer
        # self.fc = nn.Linear(hidden_dim, output_dim)
        self.sequential = nn.Sequential(
            nn.Linear(dim, dim),
            nn.LayerNorm(dim),
            nn.SiLU(),
            nn.Linear(dim, output_dim),
            nn.Softmax(dim=1),
        )

    def forward(self, x: Tensor) -> Tensor:
        batch_size, seq_length, _ = x.shape

        # Init hidden and cell states with zeros
        h = [
            torch.zeros(batch_size, self.hidden_dim).to(x.device)
            for _ in range(self.depth)
        ]
        c = [
            torch.zeros(batch_size, self.hidden_dim).to(x.device)
            for _ in range(self.depth)
        ]

        # Collect outputs for each time step
        outputs = []

        # Iterate through each time step in the sequence
        for t in range(seq_length):
            # Extract the input for the current time step
            x_t = x[:, t, :]

            # Pass through each LSTM cell
            for layer in range(self.depth):
                h[layer], c[layer] = self.cells[layer](x_t, h[layer], c[layer])
                x_t = h[layer]

            # Collect the output from the final LSTM layer
            outputs.append(h[-1].unsqueeze(1))

        # Concatenate the outputs along the time dimension
        outputs = torch.cat(outputs, dim=1)
        print(outputs.shape)
        b, s, d = outputs.shape

        # Apply the fully connected layer
        # outputs = self.sequential(outputs)
        outputs = nn.Sequential(
            nn.Linear(d, self.dim),
            nn.LayerNorm(self.dim),
            nn.SiLU(),
            nn.Linear(self.dim, self.dim),
            # nn.Softmax(dim=1),
        )(outputs)

        return outputs


# # Example usage:
# if __name__ == "__main__":
#     batch_size = 32
#     seq_length = 10
#     input_dim = 50
#     hidden_dim = 100
#     num_layers = 2
#     output_dim = 30

#     model = SimpleLSTM(input_dim, hidden_dim, num_layers, output_dim)
#     inputs = torch.randn(batch_size, seq_length, input_dim)
#     outputs = model(inputs)
#     print(outputs)  # Expected output shape: (batch_size, seq_length, output_dim)



import torch.nn.functional as F
from torch import Tensor


def mm_softmax(
    x: Tensor,
    y: Tensor,
    weight: float = 1.0,
    weight2: float = 1.0,
    temp: float = 1.0,
):
    """
    Applies softmax function to the element-wise product of two input tensors, x and y.

    Args:
        x (Tensor): The first input tensor.
        y (Tensor): The second input tensor.
        weight (float, optional): Weight multiplier for x. Defaults to 1.0.
        weight2 (float, optional): Weight multiplier for y. Defaults to 1.0.
        temp (float, optional): Temperature scaling factor. Defaults to 1.0.

    Returns:
        Tensor: The softmax output tensor.
    """
    assert x.size() == y.size(), "x and y must have the same shape"

    # Combine modalities
    combined_data = weight * x * weight2 * y

    # Apply temperature scaling
    scaled_data = combined_data / temp

    # Compute softmax on scaled combined data
    softmax = F.softmax(scaled_data, dim=-1)

    return softmax



import math
from typing import Tuple

import torch
import torch.nn as nn
import torch.nn.functional as F

# from scipy.stats import norm
from tqdm import tqdm

bnb_available = False


def get_block_absmax(
    inpt_tensor: torch.Tensor, block_size: int
) -> torch.Tensor:
    """Iterate through a flattened tensor getting the absmax scalers for each block

    Args:
        inpt_tensor: Input tensor to get scalers for
        block_size: Block size for the scanning window
    Returns:
        torch.Tensor: Tensor of scalers for each block
    """
    assert inpt_tensor.dim() == 1, "Input tensor must be flattened"
    assert (inpt_tensor.numel() % block_size) == 0, (
        "Input tensor must be divisible by block size, got"
        f" {inpt_tensor.numel()} and {block_size}"
    )

    n_blocks = inpt_tensor.numel() // block_size
    blocks = inpt_tensor.view(n_blocks, block_size)
    block_scalers = blocks.abs().max(dim=1).values
    return block_scalers


class NF4Tensor:
    """NF4Tensor class for converting a weight to the QLoRA NF4 format"""

    @classmethod
    @torch.no_grad()
    def from_tensor(
        cls,
        inpt_tensor: torch.Tensor,
        block_size: int = 64,
        scaler_block_size: int = 256,
    ):
        assert inpt_tensor.dtype == torch.bfloat16
        assert (
            inpt_tensor.numel() % block_size == 0
        ), "Input tensor must be divisible by block size"
        assert (
            inpt_tensor.dtype == torch.bfloat16
        ), "Input tensor must be bfloat16"
        device = inpt_tensor.device
        # Cache the tensor on the class def
        nf4 = torch.tensor(
            [
                -1.0000,
                -0.6962,
                -0.5251,
                -0.3949,
                -0.2844,
                -0.1848,
                -0.0911,
                0.0000,
                0.0796,
                0.1609,
                0.2461,
                0.3379,
                0.4407,
                0.5626,
                0.7230,
                1.0000,
            ],
            device=device,
            dtype=torch.bfloat16,
        )
        n_blocks = inpt_tensor.numel() // block_size
        # Double quantization
        (
            quantized_scalers,
            quantization_factor,
            scaler_mean,
        ) = cls.double_quantize_scalers(
            inpt_tensor.flatten(), block_size, scaler_block_size
        )
        quantized_data = cls.convert_to_norm_float_weight(
            inpt_tensor, n_blocks, block_size, nf4
        )
        original_shape = inpt_tensor.shape
        return cls(
            block_size,
            n_blocks,
            scaler_block_size,
            quantized_scalers,
            quantization_factor,
            scaler_mean,
            quantized_data,
            original_shape,
            nf4=nf4,
        )

    def __init__(
        self,
        block_size: int,
        n_blocks: int,
        scaler_block_size: int,
        quantized_scalers: torch.Tensor,
        quantization_factor: torch.Tensor,
        scaler_mean: torch.Tensor,
        quantized_data: torch.Tensor,
        original_shape: torch.Size,
        nf4: torch.Tensor,
    ):
        """Initialize the NF4Tensor class"""
        self.device = quantized_data.device
        self.block_size = block_size
        self.n_blocks = n_blocks
        self.scaler_block_size = scaler_block_size
        self.quantized_scalers = quantized_scalers
        self.quantization_factor = quantization_factor
        self.scaler_mean = scaler_mean
        self.quantized_data = quantized_data
        self.original_shape = original_shape
        self.nf4 = nf4

    @staticmethod
    def double_quantize_scalers(
        inpt_tensor: torch.Tensor,
        block_size: int,
        scaler_block_size: int,
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """Used to achieve the double quantization of the scalers
        We take the input tensor first calculate the absmax quantization factors for each block.
        We then find the mean of our positive absmax scalers. We subtract this mean from the scalers
        And then we calculate the absmax quantization factors for each block again. We then quantize the scalers to int8.

        Args:
            inpt_tensor: Input tensor to convert to QLoRA format, typically a weight tensor

        Returns:
            torch.Tensor: Tensor of per_block quantization factors stored in int8 format
                size: (n_blocks)
            torch.Tensor: Tensor of per_scaler_block quantization factors stored in int16 format
                size: (n_scaler_blocks)
        """
        assert inpt_tensor.dim() == 1, "Input tensor must be flattened"
        assert (inpt_tensor.numel() % scaler_block_size) == 0, (
            "Input tensor must be divisible by block size, got"
            f" {inpt_tensor.numel()} and {scaler_block_size}"
        )

        # First round of quantization
        # Produces: A tensor of size (n_blocks) of inpt_tensor.dtype
        scalers_1 = get_block_absmax(inpt_tensor, block_size)
        scalers_1_mean = scalers_1.mean()
        scalers_1 = scalers_1 - scalers_1_mean
        # Second round of quantization
        assert (
            scalers_1.numel() % scaler_block_size == 0
        ), "Number of scalers must be divisible by scaler block size"
        n_scaler_blocks = scalers_1.numel() // scaler_block_size
        scaler_blocks = scalers_1.view(n_scaler_blocks, scaler_block_size)

        scaler_absmax = get_block_absmax(scalers_1, scaler_block_size)
        scaler_absmax = scaler_absmax.unsqueeze(-1).expand(
            n_scaler_blocks, scaler_block_size
        )

        quantization_factor = 256 / (2 * scaler_absmax)
        quantized_scaler_blocks = scaler_blocks * quantization_factor
        quantized_scaler_blocks = quantized_scaler_blocks.round()
        quantized_scaler_blocks = quantized_scaler_blocks.clamp(-128, 127)

        # This is needed to make sure that quantization_factor remains a repeated view of n_scaler_blocks
        # For some reason the 127/scaler_absmax realizes n_scaler entries when only n_scaler_blocks are needed
        # The following will grab the first entry for the n_scaler_blocks which is the same across the scaler_block_size
        quantization_factor = quantization_factor[:, 0]

        return (
            quantized_scaler_blocks.flatten().to(torch.int8),
            quantization_factor.view(n_scaler_blocks),
            scalers_1_mean,
        )

    def dequantize_scalers(
        self,
        inpt_tensor: torch.Tensor,
        quantization_factor: torch.Tensor,
        scaler_block_size: int,
    ) -> torch.Tensor:
        """Used to unpack the double quantized scalers

        Args;
            inpt_tensor: Input tensor to convert to QLoRA format this is the quantized scalers in int8 format
            quantization_factor: Tensor of per_scaler_block quantization factors stored in inpt_weight.dtype
                size: (n_scaler_blocks)
            scaler_block_size: Scaler block size to use for double quantization.

        """
        assert inpt_tensor.dim() == 1, "Input tensor must be flattened"
        assert (inpt_tensor.numel() % scaler_block_size) == 0, (
            "Input tensor must be divisible by block size, got"
            f" {inpt_tensor.numel()} and {scaler_block_size}"
        )
        n_scaler_blocks = inpt_tensor.numel() // scaler_block_size
        inpt_tensor = inpt_tensor.view(n_scaler_blocks, scaler_block_size)
        dequantized = (
            inpt_tensor / quantization_factor.unsqueeze(-1)
        ).flatten().to(torch.bfloat16) + self.scaler_mean
        return dequantized

    @staticmethod
    def convert_to_norm_float_weight(
        inpt_tensor: torch.Tensor,
        n_blocks: int,
        block_size: int,
        nf4: torch.tensor,
    ) -> torch.Tensor:
        """Convert a tensor to the normalized float weight format"""
        flattened_tensor = inpt_tensor.flatten()
        #  Since we are using uint8 we will encode 2 entries per byte
        numel = inpt_tensor.numel()
        assert numel % 2 == 0, (
            "Number of elements must be even just to not have to think about"
            " the end"
        )
        # Reshape the flattened tensor into blocks of size self.block_size
        blocks = flattened_tensor.view(n_blocks, block_size)

        # Scale the blocks
        scalers = get_block_absmax(inpt_tensor.flatten(), block_size)
        scales = scalers.unsqueeze(-1).expand(n_blocks, block_size)
        scaled_blocks = blocks / scales

        # Returns a flattened tensor with each element quantized to nf4 index
        # The weird behavior comes here with how qlora vs bnb break nf4 ties.
        # Since we ust torch.min(nf4 - inpt/scale) we will always pick the smallest index
        # While bnb appears to be pick the larger index when breaking ties
        # ACTUALLYYY I think that what ever op bnb is using to get the nearest NF4 value
        # Is not consistent with torch.round. Example: input 1.1016 with abs max
        # scale of 2.2821 will get mapped to 1.25 while mine will get mapped to 0.9570
        # The difference for mine is 0.1445 and for bnb 0.1484
        quantized_blocks = NF4Tensor.quantize_tensor_nearest(
            scaled_blocks.flatten(), nf4
        )

        # Combine the quantized elements into uint8 values
        combined_blocks = quantized_blocks[::2] << 4 | quantized_blocks[1::2]

        return combined_blocks.to(torch.uint8)

    def get_original_weight(self) -> torch.Tensor:
        """Get the original weight from the normalized float weight format"""
        # since we are using uint8 we will decode 2 entries per byte
        # Shift elements down 4 and select out the bottom 4 bits
        first_elements = (self.quantized_data >> 4).to(torch.long)
        second_elements = (self.quantized_data & 0b1111).to(torch.long)

        # Dequantize every element
        dequantized_first = self.dequantize(first_elements, self.nf4)
        dequantized_second = self.dequantize(second_elements, self.nf4)

        # Build up matrix of scalers repeated for each element in the block
        # Since first and second elements make up a full block, so
        # we expand out to half the size of the full block
        scalers = self.dequantize_scalers(
            self.quantized_scalers,
            self.quantization_factor,
            self.scaler_block_size,
        )
        repeated = scalers.unsqueeze(-1).expand(
            scalers.size(0), self.block_size // 2
        )

        scaled_first = dequantized_first * repeated.flatten()
        scaled_second = dequantized_second * repeated.flatten()

        # Flip them to be vertical and them stack them together horizontally
        # Upon flattening this will interleave the elements
        scaled_first = scaled_first.unsqueeze(-1).transpose(0, 1)
        scaled_second = scaled_second.unsqueeze(-1).transpose(0, 1)
        return torch.stack([scaled_first, scaled_second], dim=-1).reshape(
            self.original_shape
        )

    @staticmethod
    def quantize_tensor_nearest(
        value: torch.float16, nf4: torch.Tensor
    ) -> torch.Tensor:
        """Quantize a float16 tensor to nf4 format to nearest and not rounded up"""
        value = value.unsqueeze(-1)  # (numel, 1)
        # Compare the value tensor with the nf4 tensor element-wise
        diff = (value - nf4).abs()
        # BnB appears to break ties by choosing the larger nf4 value
        closest_nf4 = diff.min(dim=-1).indices
        return closest_nf4

    @staticmethod
    def dequantize(value: torch.Tensor, nf4: torch.Tensor) -> torch.Tensor:
        """Dequantize a nf4 value to float16 format"""
        # return nf4.index_select(0, value)
        return nf4[value]

    def unpack(
        self,
    ) -> Tuple[
        int,
        int,
        torch.Tensor,
        torch.Tensor,
        torch.Tensor,
        torch.Tensor,
        torch.Size,
    ]:
        return (
            self.block_size,
            self.n_blocks,
            self.scaler_block_size,
            self.quantized_scalers,
            self.quantization_factor,
            self.scaler_mean,
            self.quantized_data,
            self.original_shape,
        )

    def __repr__(self):
        return (
            f"Quantized Data: {self.quantized_data}\nScalers:"
            f" {self.quantized_scalers}\n"
        )

    def __str__(self):
        return f"NF4Tensor({self.original_shape}, {self.block_size})"


class NF4TensorDebug:
    """QLoRA Weight written in a more Debug friendly manner"""

    @staticmethod
    def get_nf4(cached=True) -> torch.Tensor:
        if cached:
            return torch.tensor(
                [
                    -1.0000,
                    -0.6962,
                    -0.5251,
                    -0.3949,
                    -0.2844,
                    -0.1848,
                    -0.0911,
                    0.0000,
                    0.0796,
                    0.1609,
                    0.2461,
                    0.3379,
                    0.4407,
                    0.5626,
                    0.7230,
                    1.0000,
                ]
            )

        offset = 0.9677083
        v1 = torch.linspace(offset, 0.5, 9)[:-1].tolist()
        # v2 = [0]*(256-15)
        v3 = (torch.linspace(offset, 0.5, 8)[:-1]).tolist()
        # v = v1 + v3 + 0.0
        nkf = torch.tensor(v1 + v3 + [0.0])
        nkf = nkf.sort().values
        nkf /= nkf.max()
        return nkf

    @staticmethod
    def quantize(value: torch.float16, nkf: torch.Tensor) -> torch.Tensor:
        """Quantize a float16 value to nkf format"""
        for i in range(len(nkf)):
            if value <= nkf[i]:
                # print("value", value, "nkf", nkf[i])
                return 0 | i
        return 0 | (len(nkf) - 1)

    @staticmethod
    def quantize_nearest(
        value: torch.float16, nkf: torch.Tensor
    ) -> torch.Tensor:
        closest_index = 0
        closest_diff = abs(nkf[0] - value)
        for i in range(1, len(nkf)):
            diff = abs(nkf[i] - value)
            if diff < closest_diff:
                closest_diff = diff
                closest_index = i
        return 0 | closest_index

    @staticmethod
    def dequantize(value: torch.Tensor, nkf: torch.Tensor) -> torch.Tensor:
        """Dequantize a nkf value to float16 format"""
        # return nkf.index_select(0, value)
        return nkf[value]

    def get_scalers(
        self, inpt_tensor: torch.Tensor, block_size: int
    ) -> torch.Tensor:
        """Iterate through a flattened tensor getting the scalers for each block"""
        flattened_tensor = inpt_tensor.flatten()
        block_scalers = []
        for block_start in range(0, inpt_tensor.numel(), block_size):
            block_end = min(block_start + block_size, inpt_tensor.numel())
            block = flattened_tensor[block_start:block_end]
            block_max = block.abs().max()
            block_scalers.append(block_max)
        return torch.tensor(block_scalers)

    def __init__(self, inpt_tensor: torch.Tensor, block_size=64):
        assert inpt_tensor.dtype == torch.bfloat16
        assert (
            inpt_tensor.numel() % block_size == 0
        ), "Input tensor must be divisible by block size"
        self.block_size = block_size
        self.n_blocks = inpt_tensor.numel() // block_size
        self.scalers = self.get_scalers(inpt_tensor, self.block_size)
        self.norm_float_weight = self.get_norm_float_weight(inpt_tensor.clone())
        self.original_shape = inpt_tensor.shape

    def get_norm_float_weight(self, inpt_tensor: torch.Tensor) -> torch.Tensor:
        nkf = self.get_nf4()
        flattened_tensor = inpt_tensor.flatten()
        #  Since we are using uint8 we will encode 2 entries per byte
        numel = inpt_tensor.numel()
        assert numel % 2 == 0, (
            "Number of elements must be even just to not have to think about"
            " the end"
        )
        quantized_length = numel // 2
        quantized_tensor = torch.zeros(quantized_length, dtype=torch.uint8)
        for i in tqdm(range(len(self.scalers))):
            block_start = i * self.block_size
            block_end = min(
                block_start + self.block_size, flattened_tensor.numel()
            )
            block = flattened_tensor[block_start:block_end]
            # Scale the block
            block /= self.scalers[i]
            # We will iterate over each element in the block and quantize it
            # In groups of 2
            for j in range(0, self.block_size, 2):
                # Combine two bfloat16s via quantization to 4 bit types into a single uint8
                element_1 = self.quantize_nearest(block[j], nkf)
                element_2 = self.quantize_nearest(block[j + 1], nkf)
                combined = element_1 << 4 | element_2
                quantized_tensor[(i * self.block_size // 2) + j // 2] = combined
        return quantized_tensor

    def get_original_weight(self):
        # since we are using uint8 we will decode 2 entries per byte
        nkf = self.get_nf4()
        original_weight = torch.empty(
            2 * self.norm_float_weight.numel(), dtype=torch.bfloat16
        )
        # Scalers is a proxy for num_blocks
        for i in range(len(self.scalers)):
            block_start = i * self.block_size
            block_end = block_start + self.block_size
            block = original_weight[block_start:block_end]
            for j in range(0, self.block_size, 2):
                combined = self.norm_float_weight[
                    (i * self.block_size // 2) + j // 2
                ]
                # Shift element down 4
                element_1 = combined >> 4
                # Select out the bottom 4 bits
                element_2 = combined & 0b1111
                block[j] = (
                    self.dequantize(element_1.item(), nkf) * self.scalers[i]
                )
                block[j + 1] = (
                    self.dequantize(element_2.item(), nkf) * self.scalers[i]
                )
        return original_weight.reshape(self.original_shape)


class LinearNF4(torch.autograd.Function):
    @staticmethod
    def forward(ctx, input: torch.Tensor, weight: NF4Tensor):
        ctx.nf4_weight = weight
        return F.linear(input, weight.get_original_weight())

    @staticmethod
    def backward(ctx, grad_output):
        weight: NF4Tensor = ctx.nf4_weight
        return grad_output @ weight.get_original_weight(), None


def linear_nf4(input: torch.Tensor, weight: NF4Tensor) -> torch.Tensor:
    return LinearNF4.apply(input, weight)


def build_input_weight(embed_dim: int, device: torch.device):
    torch.manual_seed(0)
    input_weight = torch.empty(
        embed_dim, embed_dim, device=device, dtype=torch.bfloat16
    )
    input_weight.normal_(0, 1)
    return input_weight


def build_bitsandbytes_linear(input_weight: torch.Tensor, device: torch.device):
    global bnb
    if "bnb" not in globals():
        import bitsandbytes as bnb
    param = bnb.nn.Params4bit(
        input_weight, requires_grad=False, quant_type="nf4"
    ).cuda(device)
    bnb_linear = bnb.nn.LinearNF4(
        input_weight.size(0), input_weight.size(1), bias=False
    )
    bnb_linear.weight = param
    bnb_linear.to(device)
    return bnb_linear


def get_sample_inputs(
    bsz: int,
    seqlen: int,
    embed_dim: int,
    device: torch.device,
    requires_grad: bool = False,
) -> torch.Tensor:
    sample_input = torch.rand(
        bsz,
        seqlen,
        embed_dim,
        device=device,
        dtype=torch.bfloat16,
        requires_grad=requires_grad,
    )
    sample_input = sample_input.view(bsz * seqlen, embed_dim)
    return sample_input


def get_mlp_weights(
    embed_dim: int, device: torch.dtype = torch.device("cuda:0")
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """These three weights take up
    3 * (embed_dim * n_hidden) * 2 bytes of memory
    i.g. for embed_dim = 4096 and hidden_dim = 11008
    Total memory usage is 270532608 bytes or 0.27 gb
    """
    torch.manual_seed(0)

    def find_multiple(n: int, k: int) -> int:
        if n % k == 0:
            return n
        return n + k - (n % k)

    hidden_dim = 4 * embed_dim
    n_hidden = int(2 * hidden_dim / 3)
    n_hidden = find_multiple(n_hidden, 256)
    weight1 = torch.empty(
        (n_hidden, embed_dim), dtype=torch.bfloat16, device=device
    ).normal_(0, 1)
    weight2 = torch.empty(
        (n_hidden, embed_dim), dtype=torch.bfloat16, device=device
    ).normal_(0, 1)
    weight3 = torch.empty(
        (embed_dim, n_hidden), dtype=torch.bfloat16, device=device
    ).normal_(0, 1)

    return weight1, weight2, weight3


class MLP(nn.Module):
    def __init__(self, weight1, weight2, weight3) -> None:
        super().__init__()
        self.w1, self.w2, self.w3 = weight1, weight2, weight3

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = F.silu(F.linear(x, self.w1)) * F.linear(x, self.w2)
        x = F.linear(x, self.w3)
        return x


class NF4MLP(nn.Module):
    def __init__(self, weight1, weight2, weight3) -> None:
        super().__init__()
        self.w1 = NF4Tensor.from_tensor(weight1)
        self.w2 = NF4Tensor.from_tensor(weight2)
        self.w3 = NF4Tensor.from_tensor(weight3)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = F.silu(linear_nf4(x, self.w1)) * linear_nf4(x, self.w2)
        x = linear_nf4(x, self.w3)
        return x


class BnbQloraMLP(nn.Module):
    def __init__(self, weight1, weight2, weight3, device) -> None:
        super().__init__()
        self.w1 = build_bitsandbytes_linear(weight1, device)
        self.w2 = build_bitsandbytes_linear(weight2, device)
        self.w3 = build_bitsandbytes_linear(weight3, device)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = F.silu(self.w1(x)) * self.w2(x)
        x = self.w3(x)
        return x


class QloraLinear(nn.Module):
    """
    QloRA Linear Layer

    QloraLinear is a module that performs a linear transformation on the input data.

    Args:
        in_features: size of each input sample
        out_features: size of each output sample
        weight: weight tensor of shape (out_features, in_features)
        r: number of blocks to use for QLoRA
        lora_alpha: scaling factor for QLoRA
        lora_dropout: dropout to apply to the QLoRA term

    Attributes:
        weight: the learnable weights of the module of shape
            (out_features, in_features). The values are initialized from
            :math:`\\mathcal{U}(-\\sqrt{k}, \\sqrt{k})`, where :math:`k = \frac{1}{\text{in_features}}`
        lora_A: the learnable weights of the QLoRA A term of shape
            (r, in_features). The values are initialized from
            :math:`\\mathcal{U}(-\\sqrt{k}, \\sqrt{k})`, where :math:`k = \frac{1}{\text{in_features}}`
        lora_B: the learnable weights of the QLoRA B term of shape
            (out_features, r). The values are initialized to zero
        scaling: the scaling factor for the QLoRA term

    Example:
        import torch
        from zeta.nn.quant.qlora import QloraLinear
        # Convert the weight tensor to torch.bfloat16
        weight_bfloat16 = torch.rand(4096, 4096).to(torch.bfloat16)

        # Create the QloraLinear model with the correctly typed weight tensor
        model = QloraLinear(4096, 4096, weight=weight_bfloat16, r=64)

        # Convert the input tensor to torch.bfloat16
        tensor = torch.rand(4096, 4096).to(torch.bfloat16)

        # Perform a forward and backward pass
        out = model(tensor).sum()
        print(out)
        out.backward()


    """

    def __init__(
        self,
        in_features: int,
        out_features: int,
        weight: torch.Tensor,
        r: int,
        lora_alpha: int = 1,
        lora_dropout: float = 0.0,
    ) -> None:
        super().__init__()
        self.weight = NF4Tensor.from_tensor(weight)
        self.r = r
        self.lora_alpha = lora_alpha
        self.in_features = in_features
        self.out_features = out_features
        self.lora_A = nn.Parameter(weight.new_zeros((r, in_features)))
        self.lora_B = nn.Parameter(weight.new_zeros((out_features, r)))
        self.scaling = self.lora_alpha / self.r

        # Optional dropout
        if lora_dropout > 0.0:
            self.lora_dropout = nn.Dropout(p=lora_dropout)
        else:
            self.lora_dropout = lambda x: x

        self.reset_parameters()

    def reset_parameters(self):
        nn.init.kaiming_uniform_(self.lora_A, a=math.sqrt(5))
        nn.init.zeros_(self.lora_B)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        result = linear_nf4(x, self.weight)
        result2 = (
            result
            + (
                self.lora_dropout(x)
                @ self.lora_A.transpose(0, 1)
                @ self.lora_B.transpose(0, 1)
            )
            * self.scaling
        )
        return result2



from PIL import Image
from torchvision import transforms


def img_to_tensor(img: str = "pali.png", img_size: int = 256):
    """
    Convert an image to a tensor.

    Args:
        img (str): The path to the image file. Default is "pali.png".
        img_size (int): The desired size of the image. Default is 256.

    Returns:
        torch.Tensor: The image converted to a tensor.

    """
    # Load image
    image = Image.open(img)

    # Define a transforms to convert the image to a tensor and apply preprocessing
    transform = transforms.Compose(
        [
            transforms.Lambda(lambda image: image.convert("RGB")),
            transforms.Resize(
                (img_size, img_size)
            ),  # Resize the image to 256x256
            transforms.ToTensor(),  # Convert the image to a tensor,
            transforms.Normalize(
                mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]
            ),  # Normalize the pixel values
        ]
    )

    # apply transforms to the image
    x = transform(image)

    # Add batch dimension
    x = x.unsqueeze(0)

    return x



import torch
import torch.nn as nn


# Basic Block for ResNet
class BasicBlock(nn.Module):
    """BasicBlock


    Args:
        in_channels (int): Number of input channels
        out_channels (int): Number of output channels
        stride (int): Stride of the convolutional layer
        kernel_size (int): Kernel size of the convolutional layer
        padding (int): Padding of the convolutional layer
        bias (bool): Bias of the convolutional layer

    Examples:
    >>> from zeta.nn.modules.res_net import BasicBlock
    >>> import torch
    >>> x = torch.randn(5, 10)
    >>> swiglu = BasicBlock(10, 20)
    >>> swiglu(x).shape
    torch.Size([5, 10])

    """

    expansion = 1

    def __init__(
        self,
        in_channels,
        out_channels,
        stride: int = 1,
        kernel_size: int = 3,
        padding: int = 1,
        bias: bool = False,
        *args,
        **kwargs,
    ):
        super().__init__()
        self.conv1 = nn.Conv2d(
            in_channels,
            out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            bias=bias,
        )
        self.bn1 = nn.BatchNorm2d(out_channels)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = nn.Conv2d(
            out_channels,
            out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            bias=bias,
        )
        self.bn2 = nn.BatchNorm2d(out_channels)

        self.shortcut = nn.Sequential()
        if stride != 1 or in_channels != self.expansion * out_channels:
            self.shortcut = nn.Sequential(
                nn.Conv2d(
                    in_channels,
                    self.expansion * out_channels,
                    kernel_size=1,
                    stride=stride,
                    bias=bias,
                ),
                nn.BatchNorm2d(self.expansion * out_channels),
            )

    def forward(self, x: torch.Tensor):
        """Forward

        Args:
            x torch.Tensor: Input tensor

        """
        out = self.relu(self.bn1(self.conv1(x)))
        out = self.bn2(self.conv2(out))
        out += self.shortcut(x)
        out = self.relu(out)
        return out


# Full ResNet
class ResNet(nn.Module):
    """ResNet

    Args:
        block (_type_): _description_
        num_blocks (_type_): _description_
        num_classes (int): Number of classes
        kernel_size (int): Kernel size of the convolutional layer
        stride (int): Stride of the convolutional layer
        *args: Variable length argument list.
        **kwargs: Arbitrary keyword arguments.

    Examples:
    >>> from zeta.nn.modules.res_net import ResNet
    >>> import torch
    >>> x = torch.randn(5, 10)
    >>> swiglu = ResNet(10, 20)
    >>> swiglu(x).shape
    torch.Size([5, 10])


    """

    def __init__(
        self,
        block,
        num_blocks,
        num_classes: int = 1000,
        kernel_size: int = 3,
        stride: int = 2,
        *args,
        **kwargs,
    ):
        super().__init__()
        self.in_channels = 64

        self.conv1 = nn.Conv2d(
            3, 64, kernel_size=7, stride=stride, padding=3, bias=False
        )
        self.bn1 = nn.BatchNorm2d(64)
        self.relu = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool2d(
            kernel_size=kernel_size, stride=stride, padding=1
        )
        self.layer1 = self._make_layer(block, 64, num_blocks[0], stride=stride)
        self.layer2 = self._make_layer(block, 128, num_blocks[1], stride=stride)
        self.layer3 = self._make_layer(block, 256, num_blocks[2], stride=stride)
        self.layer4 = self._make_layer(block, 512, num_blocks[3], stride=stride)
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = nn.Linear(512 * block.expansion, num_classes)

    def _make_layer(self, block, out_channels, num_blocks, stride):
        """Make layer

        Args:
            block (_type_): _description_
            out_channels (_type_): _description_
            num_blocks (_type_): _description_
            stride (_type_): _description_

        Returns:
            _type_: _description_
        """
        strides = [stride] + [1] * (num_blocks - 1)
        layers = []
        for stride in strides:
            layers.append(block(self.in_channels, out_channels, stride))
            self.in_channels = out_channels * block.expansion
        return nn.Sequential(*layers)

    def forward(self, x: torch.Tensor):
        """Forward

        Args:
            x torch.Tensor: Input tensor
        """
        x = self.maxpool(self.relu(self.bn1(self.conv1(x))))
        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)
        x = self.avgpool(x)
        x = torch.flatten(x, 1)
        x = self.fc(x)
        return x


# model = ResNet(block=BasicBlock, num_blocks=[2, 2, 2, 2], num_classes=10)

# x = torch.randn(1, 3, 224, 224)

# print(model(x).shape)



from einops import rearrange
from torch import Tensor, einsum, nn

from zeta.utils.main import exists


class LinearAttention(nn.Module):
    """
    LinearAttention module performs linear attention mechanism on the input tensor.

    Args:
        dim (int): The dimension of the input tensor.
        heads (int, optional): The number of attention heads. Defaults to 4.
        dim_head (int, optional): The dimension of each attention head. Defaults to 64.
        dropout (float, optional): The dropout probability. Defaults to 0.0.

    Returns:
        Tensor: The output tensor after linear attention mechanism.


    Example::
        >>> import torch
        >>> from zeta.nn.attention import LinearAttention
        >>> x = torch.randn(1, 32, 64)
        >>> attn = LinearAttention(64)
        >>> out = attn(x)
        >>> out.shape
        torch.Size([1, 32, 64])
    """

    def __init__(
        self,
        dim: int,
        heads: int = 4,
        dim_head: int = 64,
        dropout: float = 0.0,
        *args,
        **kwargs,
    ):
        super().__init__()
        self.dim = dim
        self.heads = heads
        self.dim_head = dim_head
        self.dropout = dropout

        inner_dim = heads * dim_head
        self.scale = dim_head**-0.5

        # Linear projection layers
        self.to_qkv = nn.Linear(dim, inner_dim * 3, bias=False)
        self.to_out = nn.Sequential(
            nn.Linear(inner_dim, dim), nn.Dropout(dropout)
        )

    def forward(self, x: Tensor, mask: Tensor = None):
        """
        Perform forward pass of the LinearAttention module.

        Args:
            x (Tensor): The input tensor.
            mask (Tensor, optional): The mask tensor. Defaults to None.

        Returns:
            Tensor: The output tensor after linear attention mechanism.
        """
        h = self.heads
        q, k, v = self.to_qkv(x).chunk(3, dim=-1)
        q, k, v = map(
            lambda t: rearrange(t, "b n (h d) -> (b h) n d", h=h), (q, k, v)
        )

        q = q * self.scale
        q, k = q.softmax(dim=-1), k.softmax(dim=-2)

        if exists(mask):
            k.masked_fill(mask, 0.0)

        context = einsum("b n d, b n e -> b d e", q, k)
        out = einsum("b d e, b n d -> b n e", context, v)
        out = rearrange(out, "(b h) n d -> b n (h d)", h=h)
        return self.to_out(out)



import enum
import logging
from typing import List, Tuple, Union

import torch
from einops import rearrange
from torch import Tensor

logger = logging.getLogger(__name__)


class NewtonConvergenceFlag(enum.Enum):
    REACHED_MAX_ITERS = 0
    CONVERGED = 1


class RootInvMethod(enum.Enum):
    EIGEN = 0
    NEWTON = 1


def check_diagonal(A: Tensor) -> Tensor:
    """Checks if symmetric matrix is diagonal."""

    A_shape = A.shape
    if len(A_shape) != 2:
        raise ValueError("Matrix is not 2-dimensional!")

    m, n = A_shape
    if m != n:
        raise ValueError("Matrix is not square!")

    return ~torch.any(A.reshape(-1)[:-1].reshape(m - 1, n + 1)[:, 1:].bool())


def matrix_inverse_root(
    A: Tensor,
    root: int,
    epsilon: float = 0.0,
    exponent_multiplier: float = 1.0,
    root_inv_method: RootInvMethod = RootInvMethod.EIGEN,
    max_iterations: int = 1000,
    tolerance: float = 1e-6,
    is_diagonal: Union[Tensor, bool] = False,
    retry_double_precision: bool = True,
) -> Tensor:
    """Computes matrix root inverse of square symmetric positive definite matrix.

    Args:
        A (Tensor): Square matrix of interest.
        root (int): Root of interest. Any natural number.
        epsilon (float): Adds epsilon * I to matrix before taking matrix root. (Default: 0.0)
        exponent_multiplier (float): exponent multiplier in the eigen method (Default: 1.0)
        root_inv_method (RootInvMethod): Specifies method to use to compute root inverse. (Default: RootInvMethod.EIGEN)
        max_iterations (int): Maximum number of iterations for coupled Newton iteration. (Default: 1000)
        tolerance (float): Tolerance for computing root inverse using coupled Newton iteration. (Default: 1e-6)
        is_diagonal (Tensor, bool): Flag for whether or not matrix is diagonal. If so, will compute root inverse by computing
            root inverse of diagonal entries. (Default: False)
        retry_double_precision (bool): Flag for re-trying eigendecomposition with higher precision if lower precision fails due
            to CuSOLVER failure. (Default: True)

    Returns:
        X (Tensor): Inverse root of matrix A.

    """

    # check if matrix is scalar
    if torch.numel(A) == 1:
        alpha = torch.as_tensor(-exponent_multiplier / root)
        return (A + epsilon) ** alpha

    # check matrix shape
    if len(A.shape) != 2:
        raise ValueError("Matrix is not 2-dimensional!")
    elif A.shape[0] != A.shape[1]:
        raise ValueError("Matrix is not square!")

    if is_diagonal:
        X = matrix_root_diagonal(
            A=A,
            root=root,
            epsilon=epsilon,
            inverse=True,
            exponent_multiplier=exponent_multiplier,
            return_full_matrix=True,
        )
    elif root_inv_method == RootInvMethod.EIGEN:
        X, _, _ = _matrix_root_eigen(
            A=A,
            root=root,
            epsilon=epsilon,
            inverse=True,
            exponent_multiplier=exponent_multiplier,
            retry_double_precision=retry_double_precision,
        )
    elif root_inv_method == RootInvMethod.NEWTON:
        if exponent_multiplier != 1.0:
            raise ValueError(
                f"Exponent multiplier {exponent_multiplier} must be equal to 1"
                " to use coupled inverse Newton iteration!"
            )

        X, _, termination_flag, _, _ = _matrix_inverse_root_newton(
            A=A,
            root=root,
            epsilon=epsilon,
            max_iterations=max_iterations,
            tolerance=tolerance,
        )
        if termination_flag == NewtonConvergenceFlag.REACHED_MAX_ITERS:
            logging.warning(
                "Newton did not converge and reached maximum number of"
                " iterations!"
            )
    else:
        raise NotImplementedError(
            "Root inverse method is not implemented! Specified root inverse"
            " method is " + str(root_inv_method) + "."
        )

    return X


def matrix_root_diagonal(
    A: Tensor,
    root: int,
    epsilon: float = 0.0,
    inverse: bool = True,
    exponent_multiplier: float = 1.0,
    return_full_matrix: bool = False,
) -> Tensor:
    """Computes matrix inverse root for a diagonal matrix by taking inverse square root of diagonal entries.

    Args:
        A (Tensor): One- or two-dimensional tensor containing either the diagonal entries of the matrix or a diagonal matrix.
        root (int): Root of interest. Any natural number.
        epsilon (float): Adds epsilon * I to matrix before taking matrix root. (Default: 0.0)
        inverse (bool): Returns inverse root matrix. (Default: True)
        return_full_matrix (bool): Returns full matrix by taking torch.diag of diagonal entries. (bool: False)

    Returns:
        X (Tensor): Inverse root of diagonal entries.

    """

    # check order of tensor
    order = len(A.shape)
    if order == 2:
        A = torch.diag(A)
    elif order > 2:
        raise ValueError("Matrix is not 2-dimensional!")

    # check if root is positive integer
    if root <= 0:
        raise ValueError(f"Root {root} should be positive!")

    # compute matrix power
    alpha = exponent_multiplier / root
    if inverse:
        alpha = -alpha

    X = (A + epsilon).pow(alpha)
    return torch.diag(X) if return_full_matrix else X


def _matrix_root_eigen(
    A: Tensor,
    root: int,
    epsilon: float = 0.0,
    inverse: bool = True,
    exponent_multiplier: float = 1.0,
    make_positive_semidefinite: bool = True,
    retry_double_precision: bool = True,
) -> Tuple[Tensor, Tensor, Tensor]:
    """Compute matrix (inverse) root using eigendecomposition of symmetric positive (semi-)definite matrix.

            A = Q L Q^T => A^{1/r} = Q L^{1/r} Q^T OR A^{-1/r} = Q L^{-1/r} Q^T

    Assumes matrix A is symmetric.

    Args:
        A (Tensor): Square matrix of interest.
        root (int): Root of interest. Any natural number.
        epsilon (float): Adds epsilon * I to matrix before taking matrix root. (Default: 0.0)
        inverse (bool): Returns inverse root matrix. (Default: True)
        exponent_multiplier (float): exponent multiplier in the eigen method (Default: 1.0)
        make_positive_semidefinite (bool): Perturbs matrix eigenvalues to ensure it is numerically positive semi-definite. (Default: True)
        retry_double_precision (bool): Flag for re-trying eigendecomposition with higher precision if lower precision fails due
            to CuSOLVER failure. (Default: True)

    Returns:
        X (Tensor): (Inverse) root of matrix. Same dimensions as A.
        L (Tensor): Eigenvalues of A.
        Q (Tensor): Orthogonal matrix consisting of eigenvectors of A.

    """

    # check if root is positive integer
    if root <= 0:
        raise ValueError(f"Root {root} should be positive!")

    # compute matrix power
    alpha = exponent_multiplier / root
    if inverse:
        alpha = -alpha

    # compute eigendecomposition and compute minimum eigenvalue
    try:
        L, Q = torch.linalg.eigh(A)

    except Exception as exception:
        if retry_double_precision and A.dtype != torch.float64:
            logger.warning(
                f"Failed to compute eigendecomposition in {A.dtype} precision"
                f" with exception {exception}! Retrying in double precision..."
            )
            L, Q = torch.linalg.eigh(A.double())
        else:
            raise exception

    lambda_min = torch.min(L)

    # make eigenvalues >= 0 (if necessary)
    if make_positive_semidefinite:
        L += -torch.minimum(lambda_min, torch.as_tensor(0.0))

    # add epsilon
    L += epsilon

    # compute inverse preconditioner
    X = Q * L.pow(alpha).unsqueeze(0) @ Q.T

    return X, L, Q


def _matrix_inverse_root_newton(
    A,
    root: int,
    epsilon: float = 0.0,
    max_iterations: int = 1000,
    tolerance: float = 1e-6,
) -> Tuple[Tensor, Tensor, NewtonConvergenceFlag, int, Tensor]:
    """Compute matrix inverse root using coupled inverse Newton iteration.

        alpha <- -1 / p
        X <- 1/c * I
        M <- 1/c^p * A
        repeat until convergence
            M' <- (1 - alpha) * I + alpha * M
            X <- X * M'
            M <- M'^p * M

    where c = (2 |A|_F / (p + 1))^{1/p}. This ensures that |A|_2 <= |A|_F < (p + 1) c^p, which guarantees convergence.
    We will instead use z = (p + 1) / (2 * |A|_F).

    NOTE: Exponent multiplier not compatible with coupled inverse Newton iteration!

    Args:
        A (Tensor): Matrix of interest.
        root (int): Root of interest. Any natural number.
        epsilon (float): Adds epsilon * I to matrix before taking matrix root. (Default: 0.0)
        max_iterations (int): Maximum number of iterations. (Default: 1000)
        tolerance (float): Tolerance. (Default: 1e-6)

    Returns:
        A_root (Tensor): Inverse square root of matrix.
        M (Tensor): Coupled matrix.
        termination_flag (NewtonConvergenceFlag): Specifies convergence.
        iteration (int): Number of iterations.
        error (Tensor): Final error between M and I.

    """

    # initialize iteration, dimension, and alpha
    iteration = 0
    dim = A.shape[0]
    alpha = -1 / root
    identity = torch.eye(dim, dtype=A.dtype, device=A.device)

    # add regularization
    A.add_(identity, alpha=epsilon)

    # initialize matrices
    A_nrm = torch.linalg.norm(A)
    z = (root + 1) / (2 * A_nrm)
    X = z ** (-alpha) * identity
    M = z * A
    # pyre-fixme[6]: For 1st param expected `Tensor` but got `float`.
    error = torch.dist(M, identity, p=torch.inf)

    # main for loop
    while error > tolerance and iteration < max_iterations:
        iteration += 1
        M_p = M.mul(alpha).add_(identity, alpha=(1 - alpha))
        X = X @ M_p
        M = torch.linalg.matrix_power(M_p, root) @ M
        error = torch.dist(M, identity, p=torch.inf)

    # determine convergence flag
    termination_flag = (
        NewtonConvergenceFlag.CONVERGED
        if error <= tolerance
        else NewtonConvergenceFlag.REACHED_MAX_ITERS
    )

    return X, M, termination_flag, iteration, error


def compute_matrix_root_inverse_residuals(
    A: Tensor,
    X_hat: Tensor,
    root: int,
    epsilon: float,
    exponent_multiplier: float,
) -> Tuple[Tensor, Tensor]:
    """Compute residual of matrix root inverse for debugging purposes.

        relative error    = ||X - X_hat||_inf / ||X||_inf
        relative residual = ||A X^r - I||_inf

    Args:
        A (Tensor): Matrix of interest.
        X (Tensor): Computed matrix root inverse.
        root (int): Root of interest.
        epsilon (float): Adds epsilon * I to matrix.
        exponent_multiplier (float): Exponent multiplier to be multiplied to the numerator of the inverse root.

    Returns:
        absolute_error (Tensor): absolute error of matrix root inverse
        relative_error (Tensor): relative error of matrix root inverse
        residual (Tensor): residual of matrix root inverse

    """

    # check shape of matrix
    if len(A.shape) != 2:
        raise ValueError("Matrix is not 2-dimensional!")
    elif A.shape[0] != A.shape[1]:
        raise ValueError("Matrix is not square!")
    elif A.shape != X_hat.shape:
        raise ValueError("Matrix shapes do not match!")

    # compute error by comparing against double precision
    X = matrix_inverse_root(
        A.double(),
        root,
        epsilon=epsilon,
        exponent_multiplier=exponent_multiplier,
    )
    relative_error = torch.dist(X, X_hat, p=torch.inf) / torch.norm(
        X, p=torch.inf
    )

    # compute residual
    if exponent_multiplier == 1.0:
        X_invr = torch.linalg.matrix_power(X_hat.double(), n=-root)
    else:
        X_invr, _, _ = _matrix_root_eigen(
            X_hat.double(),
            root=1,
            epsilon=0.0,
            inverse=True,
            make_positive_semidefinite=True,
            exponent_multiplier=root / exponent_multiplier,
        )

    A_reg = A.double() + epsilon * torch.eye(
        A.shape[0], dtype=torch.float64, device=A.device
    )
    relative_residual = torch.dist(X_invr, A_reg, p=torch.inf) / torch.norm(
        A_reg, p=torch.inf
    )

    return relative_error, relative_residual


def merge_small_dims(tensor_shape: List[int], threshold: int) -> List[int]:
    """
    Reshapes tensor by merging small dimenions

    Args:
    tensor_shape (List[int]):  the shape of the tensor
    threshold(int) threshold on the maximum size of each dimension

    Returns:
        new_tensor_shape (List[int]) New tensor shape
    """

    new_tensor_shape = [tensor_shape[0]]
    for next_tensor_shape in tensor_shape[1:]:
        new_dimension = new_tensor_shape[-1] * next_tensor_shape
        if (
            new_tensor_shape[-1] == 1
            or next_tensor_shape == 1
            or new_dimension <= threshold
        ):
            new_tensor_shape[-1] = new_dimension
        else:
            new_tensor_shape.append(next_tensor_shape)

    return new_tensor_shape


def multi_dim_split(
    tensor: Tensor,
    splits: List[int],
) -> List[Tensor]:
    """
    Chunks tensor across multiple dimenions based on splits

    Args;
        tensor(tensor): gradient or tensor to split
        splits (List[int]) List of sizes for each block or chunk along each dimension

    Returns:
        split_grad(List[Tensor]): List of tensors
    """
    split_tensors = [tensor]
    for dim, split in enumerate(splits):
        split_tensors = [
            s for t in split_tensors for s in torch.split(t, split, dim=dim)
        ]
    return split_tensors


def multi_dim_cat(split_tensors: List[Tensor], num_splits: List[int]) -> Tensor:
    """Concatenates multiple tensors to form single tensor across multiple dimensions.

    Args:
        split_tensor (List[Tensor]): List of tensor splits or blocks.
        num_splits (List[int]): Number of splits/blocks.

    Returns:
        merged_tensor (Tensor): Merged tensor.

    """
    merged_tensor = split_tensors
    for dim, split in reversed(list(enumerate(num_splits))):
        if split > 0:
            merged_tensor = [
                torch.cat(merged_tensor[i : i + split], dim=dim)
                for i in range(0, len(merged_tensor), split)
            ]
    assert len(merged_tensor) == 1
    return merged_tensor[0]


def img_transpose(x):
    return rearrange(x, "b c h w -> b h w c")


def img_transpose_2daxis(x):
    return rearrange(x, "h w c -> w h c")


def img_composition_axis(x):
    return rearrange(x, "b h w c -> (b h) w c")


def img_compose_bw(x):
    return rearrange(x, "b h w c -> h (b w) c")


# decomposition of axis => inverse process, which represents an xis as a combination of new axis
# b1=2 is to decompose to 6 to b1=2 and b2=3
def img_decompose(x):
    return rearrange(x, "(b1 b2) h w c -> b1 b2 h w c", b1=2).shape


def img_compose_decompose(x):
    return rearrange(x, "(b1 b2) h w c -> (b1 h) (b2 w) c", b1=2)


# b1 is merged with width and b2 with height
def img_comp_decomp_merge(x):
    return rearrange(x, "(b1 b2) h w c -> (b2 h) (b1 w) c", b1=2)


# move part of width dimension to height
# width to height as image width shrunk by 2 and height doubled
def img_width_to_height(x):
    return rearrange(x, "b h (w w2) c -> (h w2) (b w) c", w2=2)


# order of axes
def img_order_of_axes(x):
    return rearrange(x, "b h w c -> h (b w) c")


# for each batch and for each pair of channels we sum over h and w
def gram_matrix_new(y):
    b, ch, h, w = y.shape
    return torch.einsum(
        "bchw,bdhw->bcd",
        [y, y],
    ) / (h * w)


# channel shuffle from shufflenet
def channel_shuffle_new(x, groups):
    return rearrange(
        x,
        "b (c1 c2) h w -> b (c2 c1) h w",
        c1=groups,
    )


# GLOW depth to space
def unsqueeze_2d_new(input, factor=2):
    return rearrange(
        input, "b (c h2 w2) h w -> b c (h h2) (w w2)", h2=factor, w2=factor
    )


def squeeze_2d_new(input, factor=2):
    return rearrange(
        input, "b c (h h2) (w w2) -> b (c h2 w2) h w", h2=factor, w2=factor
    )



import random

import torch
from sumtree import SumTree


class PrioritizedReplayBuffer:
    def __init__(
        self,
        state_size,
        action_size,
        buffer_size,
        device,
        eps=1e-2,
        alpha=0.1,
        beta=0.1,
    ):
        """
        Initializes a PrioritizedReplayBuffer object.

        Args:
            state_size (int): The size of the state space.
            action_size (int): The size of the action space.
            buffer_size (int): The maximum capacity of the buffer.
            device (torch.device): The device to store the tensors on.
            eps (float, optional): A small constant added to the priorities to ensure non-zero probabilities. Defaults to 1e-2.
            alpha (float, optional): The exponent used to compute the priority weights. Defaults to 0.1.
            beta (float, optional): The exponent used to compute the importance sampling weights. Defaults to 0.1.
        """
        self.tree = SumTree(size=buffer_size)

        self.eps = eps
        self.alpha = alpha
        self.beta = beta
        self.max_priority = 1.0

        self.state = torch.empty(buffer_size, state_size, dtype=torch.float)
        self.action = torch.empty(buffer_size, action_size, dtype=torch.float)
        self.reward = torch.empty(buffer_size, dtype=torch.float)
        self.next_state = torch.empty(
            buffer_size, state_size, dtype=torch.float
        )
        self.done = torch.empty(buffer_size, dtype=torch.uint8)

        self.count = 0
        self.real_size = 0
        self.size = buffer_size

        # device
        self.device = device

    def add(self, transition):
        """
        Adds a transition to the replay buffer.

        Args:
            transition (tuple): A tuple containing the state, action, reward, next_state, and done flag.
        """
        state, action, reward, next_state, done = transition

        self.tree.add(self.max_priority, self.count)

        self.state[self.count] = torch.as_tensor(state)
        self.action[self.count] = torch.as_tensor(action)
        self.reward[self.count] = torch.as_tensor(reward)
        self.next_state[self.count] = torch.as_tensor(next_state)
        self.done[self.count] = torch.as_tensor(done)

        self.count = (self.count + 1) % self.size
        self.real_size = min(self.size, self.real_size + 1)

    def sample(self, batch_size):
        """
        Samples a batch of transitions from the replay buffer.

        Args:
            batch_size (int): The size of the batch to sample.

        Returns:
            tuple: A tuple containing the batch of transitions, importance sampling weights, and tree indices.
        """
        assert (
            self.real_size >= batch_size
        ), "buffer contains fewer samples than batch size"

        sample_idxs, tree_idxs = [], []
        priorities = torch.empty(batch_size, 1, dtype=torch.float)

        segment = self.tree.total / batch_size
        for i in range(batch_size):
            a, b = segment * i, segment * (i + 1)

            cumsum = random.uniform(a, b)

            tree_idx, priority, sample_idx = self.tree.get(cumsum)

            priorities[i] = priority
            tree_idxs.append(tree_idx)
            sample_idxs.append(sample_idx)

        probs = priorities / self.tree.total

        weights = (self.real_size * probs) ** -self.beta

        weights = weights / weights.max()
        batch = (
            self.state[sample_idxs].to(self.device),
            self.action[sample_idxs].to(self.device),
            self.reward[sample_idxs].to(self.device),
            self.next_state[sample_idxs].to(self.device),
            self.done[sample_idxs].to(self.device),
        )
        return batch, weights, tree_idxs

    def update_priorities(self, data_idxs, priorities):
        """
        Updates the priorities of the transitions in the replay buffer.

        Args:
            data_idxs (list): A list of indices corresponding to the transitions in the replay buffer.
            priorities (torch.Tensor or numpy.ndarray): The updated priorities for the corresponding transitions.
        """
        if isinstance(priorities, torch.Tensor):
            priorities = priorities.detach().cpu().numpy()

        for data_idx, priority in zip(data_idxs, priorities):
            priority = (priority + self.eps) ** self.alpha
            self.tree.update(data_idx, priority)
            self.max_priority = max(self.max_priority, priority)



import torch
from torch import nn, Tensor


class LinearizedAttention(nn.Module):
    def __init__(
        self,
        dim: int,
        heads: int = 8,
        seqlen: int = 1000,
        groups: int = 1,
        mask_on: bool = False,
        *args,
        **kwargs,
    ):
        """
        Linearized Attention module.

        Args:
            dim (int): Dimension of the input tensor.
            heads (int): Number of attention heads.
            seqlen (int): Length of the input sequence.
            groups (int, optional): Number of groups for group normalization. Defaults to 1.
        """
        super().__init__()
        self.dim = dim
        self.heads = heads
        self.seqlen = seqlen
        self.groups = groups
        self.mask_on = mask_on

        # Projection
        self.proj = nn.Linear(dim, dim)

        # RELU
        self.act = nn.ReLU()

        # Groupnorm
        self.norm = nn.GroupNorm(groups, dim)

        # Mask Tensor
        self.mask_tensor = torch.zeros(1, seqlen).bool()

    def forward(self, x: Tensor, mask: bool = None) -> Tensor:
        """
        Forward pass of the LinearizedAttention module.

        Args:
            x (Tensor): Input tensor.

        Returns:
            Tensor: Output tensor after applying LinearizedAttention.
        """
        b, s, d = x.shape
        q = self.proj(x)
        k = self.proj(x)
        v = self.proj(x)

        # Projected again
        q_p = self.proj(q)
        q_k = self.proj(k)

        # Apply the relu
        q_acted = self.act(q_p)
        k_acted = self.act(q_k)

        # Groupnorm
        output = nn.GroupNorm(self.groups, s)(q_acted + k_acted + v)

        # Apply mask
        if mask is not None:
            if self.mask_on is True:
                mask = self.mask_tensor
            else:
                output = output.masked_fill(mask.unsqueeze(-1), float("-inf"))
                print(output.shape)

        return output


# x = torch.randn(1, 10, 20)
# model = LinearizedAttention(20, 8, mask_on=True)
# print(model(x))
# # torch.Size([1, 10, 20])
